from typing import List
from .data_explorer import DataExplorer


class ClinicalCriteria:
    """
    A class representing clinical criteria.

    - **Attributes** ::

        :param clinical_criteria_dict: dictionary containing metadata about clinical criteria

        :Example
                dict = {"disease_site": "Lung",
                "protocol_name": "Lung_Default_2Gy_30Fr",
                "pres_per_fraction_gy": 2,
                "num_of_fractions": 30,
                "criteria": [
                    {
                      "name": "max_dose",
                      "parameters": {
                        "struct": "GTV"
                      }}

    - **Methods** ::
        :add_criterion(criterion:str, parameters:dict, constraints:dict)
        :modify_criterion(criterion:str, parameters:dict, constraints:dict)

    """

    def __init__(self, data: DataExplorer, protocol_name: str = None, protocol_type: str = 'Default'):
        """

        :param clinical_criteria: dictionary containing information about clinical criteria

        """
        clinical_criteria_dict = data.load_config_clinical_criteria(protocol_name, protocol_type=protocol_type)
        self.clinical_criteria_dict = clinical_criteria_dict

    def get_prescription(self) -> float:
        """

        :return: prescription in Gy
        """
        pres = self.clinical_criteria_dict['pres_per_fraction_gy'] * \
               self.clinical_criteria_dict[
                   'num_of_fractions']
        return pres

    def get_num_of_fractions(self) -> float:
        """

        :return: number of fractions to be delivered
        """
        return self.clinical_criteria_dict['num_of_fractions']

    def add_criterion(self, criterion: str, parameters: dict, constraints: dict) -> None:
        """
        Add criterion to the clinical criteria dictionary

        :param criterion: criterion name. e.g. max_dose
        :param parameters: parameters dictionary e.g. parameters = {'struct':'PTV'}
        :param constraints: constraints dictionary e.g. constraints = {'limit_dose_gy':66, 'goal_dose_gy':60}
        :return: add the criteria to clinical criteria dictionary

        """

        self.clinical_criteria_dict['criteria'].append({'name': criterion})
        self.clinical_criteria_dict['criteria'][-1]['parameters'] = parameters
        self.clinical_criteria_dict['criteria'][-1]['constraints'] = constraints

    def modify_criterion(self, criterion: str, parameters: dict, constraints: dict) -> None:
        """

        Modify the criterion in clinical criteria dictionary

        :param criterion: criterion name. e.g. max_dose
        :param parameters: parameters dictionary e.g. parameters = {'struct':'PTV'}
        :param constraints: constraints dictionary e.g. constraints = {'limit_dose_gy':66, 'goal_dose_gy':60}
        :return: add the criteria to clinical criteria dictionary

        """
        criteria = self.clinical_criteria_dict['criteria']
        ind = [criteria[i]['name'] for i in range(len(criteria)) if criteria[i]['name'] == criterion and
               criteria[i]['parameters'] == parameters]
        if len(ind) > 0:

            criteria[ind[0]]['constraints'] = constraints
        else:
            raise Exception('No criteria  for name {}  and parameters {}'.format(criterion, parameters))

    @staticmethod
    def create_criterion(criterion: str, parameters: dict, constraints: dict):
        """
        Create criterion and return list

        :param criterion: criterion name. e.g. max_dose
        :param parameters: parameters dictionary e.g. parameters = {'struct':'PTV'}
        :param constraints: constraints dictionary e.g. constraints = {'limit_dose_gy':66, 'goal_dose_gy':60}
        :return: add the criteria to clinical criteria dictionary

        """

        criterion = [{'name': criterion, 'parameters': parameters, 'constraints': constraints}]
        return criterion

    def get_criteria(self, name: str = None) -> List[dict]:
        """
        Returns all the clinical criteria
        :return:
        """
        all_criteria = []
        if name is None:
            all_criteria = self.clinical_criteria_dict['criteria']
        elif name == 'max_dose':
            criteria = self.clinical_criteria_dict['criteria']
            ind = [i for i in range(len(criteria)) if criteria[i]['name'] == name]
            if len(ind) > 0:
                all_criteria = [criteria[i] for i in ind]
        elif name == 'mean_dose':
            criteria = self.clinical_criteria_dict['criteria']
            ind = [i for i in range(len(criteria)) if criteria[i]['name'] == name]
            if len(ind) > 0:
                all_criteria = [criteria[i] for i in ind]

        if isinstance(all_criteria, dict):
            all_criteria = [all_criteria]
        return all_criteria



from scipy import interpolate
import numpy as np
import pandas as pd
import webbrowser
from typing import List, Union
import os

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from .plan import Plan
from .clinical_criteria import ClinicalCriteria
from tabulate import tabulate


class Evaluation:
    """
    Class for evaluating different metrics for the plan

    - **Methods** ::
        :display_clinical_criteria(my_plan, sol)
            display plan values corresponding to given clinical criteria
        :get_dose(sol, struct, volume_per)
            get dose at the given volume in percentage
        :get_volume(sol: struct dose_value_gy)
            Get volume at dose_1d value in Gy


    """

    @staticmethod
    def display_clinical_criteria(my_plan: Plan, sol: Union[dict, List[dict]], html_file_name='temp.html',
                                  sol_names: List[str] = None, clinical_criteria: ClinicalCriteria = None,
                                  return_df: bool = False, in_browser: bool = False):
        """
        Visualization the plan metrics for clinical criteria in browser.
        It evaluate the plan by comparing the metrics against required criteria.

        If plan value is green color. It meets all the Limits and Goals
        If plan value is yellow color. It meets limits but not goals
        If plan value is red color. It violates both limit and goals

        :param my_plan: object of class Plan
        :param sol: optimal solution dictionary
        :param html_file_name:  name of the html file to be launched in browser
        :param sol_names: Default to Plan Value. column names for the plan evaluation
        :param clinical_criteria: clinical criteria to be evaluated
        :param return_df: return df instead of visualization
        :param in_browser: display table in browser
        :return: plan metrics in browser
        """

        # convert clinical criteria in dataframe
        if clinical_criteria is None:
            clinical_criteria = my_plan.clinical_criteria
        df = pd.DataFrame.from_dict(clinical_criteria.clinical_criteria_dict['criteria'])
        if isinstance(sol, dict):
            sol = [sol]
        if sol_names is None:
            if len(sol) > 1:
                sol_names = ['Plan Value ' + str(i) for i in range(len(sol))]
            else:
                sol_names = ['Plan Value']
        for p, s in enumerate(sol):
            dose_1d = s['inf_matrix'].A @ (s['optimal_intensity'] * my_plan.get_num_of_fractions())
            for ind in range(len(df)):  # Loop through the clinical criteria
                if 'structure_def' in df.parameters[ind]:
                    del df.parameters[ind]['structure_def']
                if df.name[ind] == 'max_dose':
                    struct = df.parameters[ind]['structure_name']
                    if struct in my_plan.structures.get_structures():
                        max_dose = Evaluation.get_max_dose(s, dose_1d=dose_1d, struct=struct)  # get max dose_1d
                        if 'limit_dose_gy' in df.constraints[ind] or 'goal_dose_gy' in df.constraints[ind]:
                            df.at[ind, sol_names[p]] = max_dose
                        elif 'limit_dose_perc' in df.constraints[ind] or 'goal_dose_perc' in df.constraints[ind]:
                            df.at[ind, sol_names[p]] = max_dose / my_plan.get_prescription() * 100
                if df.name[ind] == 'mean_dose':
                    struct = df.parameters[ind]['structure_name']
                    if struct in my_plan.structures.get_structures():
                        mean_dose = Evaluation.get_mean_dose(s, dose_1d=dose_1d, struct=struct)
                        df.at[ind, sol_names[p]] = mean_dose
                if df.name[ind] == 'dose_volume_V':
                    struct = df.parameters[ind]['structure_name']
                    if struct in my_plan.structures.get_structures():
                        if 'limit_volume_perc' in df.constraints[ind] or 'goal_volume_perc' in df.constraints[ind]:
                            dose = df.parameters[ind]['dose_gy']
                            volume = Evaluation.get_volume(s, dose_1d=dose_1d, struct=struct, dose_value_gy=dose)
                            df.at[ind, sol_names[p]] = volume
                        elif 'limit_volume_cc' in df.constraints[ind] or 'goal_volume_cc' in df.constraints[ind]:
                            dose = df.parameters[ind]['dose_gy']
                            volume = Evaluation.get_volume(s, dose_1d=dose_1d, struct=struct, dose_value_gy=dose)
                            vol_cc = my_plan.structures.get_volume_cc(structure_name=struct) * volume / 100
                            df.at[ind, sol_names[p]] = vol_cc
        pd.set_option("display.precision", 3)
        df.dropna(axis=0, inplace=True)  # remove structures which are not present
        df.reset_index(drop=True, inplace=True)  # reset the index

        def color_plan_value(row):

            highlight_red = 'background-color: red;'
            highlight_green = 'background-color: green;'
            highlight_orange = 'background-color: orange;'
            default = ''

            def matching_keys(dictionary, search_string):  # method to find the key of only part of the key match
                """

                :param dictionary: dictionary to search the string
                :param search_string: string to be searched in dictionary
                :return: full key if found search string found else empty string
                """
                get_key = None
                for key, val in dictionary.items():
                    if search_string in key:
                        get_key = key
                if get_key is not None:
                    return get_key
                else:
                    return ''

            limit_key = matching_keys(row['constraints'], 'limit')
            goal_key = matching_keys(row['constraints'], 'goal')

            row_color = len(row) * [default]  # default color for all rows initially
            # must return one string per cell in this row
            for i in range(len(row) - 1):
                if limit_key in row['constraints']:
                    if not (pd.isnull(row[i])):
                        if row[i] > row['constraints'][limit_key] + 0.0001:  # added epsilon to avoid minor differences
                            row_color[i] = highlight_red  # make plan value in red
                        else:
                            row_color[i] = highlight_green  # make plan value in red
                if goal_key in row['constraints']:
                    if not (pd.isnull(row[i])):
                        if row[i] > row['constraints'][goal_key] + 0.0001:
                            row_color[i] = highlight_orange  # make plan value in red
                        else:
                            row_color[i] = highlight_green  # make plan value in red
            return row_color

        sol_names.append('constraints')
        styled_df = df.style.apply(color_plan_value, subset=sol_names, axis=1)  # apply
        # color to dataframe using df.style method
        if return_df:
            return styled_df
        if in_browser:
            html = styled_df.render()  # render to html
            html_string = '''
                                                <html>
                                                  <head><title>Portpy Clinical Criteria Evaluation</title></head>
                                                  <style> 
                                                    table, th, td {{font-size:10pt; border:1px solid black; border-collapse:collapse; text-align:left;}}
                                                    th, td {{padding: 5px;}}
                                                  </style>
                                                  <body>
                                                  <h1> Clinical Criteria</h1>
                                                  <h4 style="color: green">Meets limit and goal</h4>
                                                  <h4 style="color: orange">Meets limit but not goal</h4>
                                                  <h4 style="color: red">Violate both limit and goal</h4>
                                                    {table}
                                                  </body>
                                                </html>.
                                                '''
            with open(html_file_name, 'w') as f:
                f.write(html_string.format(table=html))
            webbrowser.open('file://' + os.path.realpath(html_file_name))

        else:
            if Evaluation.is_notebook():
                from IPython.display import display
                with pd.option_context('display.max_rows', None,
                                       'display.max_columns', None,
                                       'display.precision', 3,
                                       ):

                    display(styled_df)
            else:
                print(tabulate(df, headers='keys', tablefmt='psql'))  # print in console using tabulate

    @staticmethod
    def get_dose(sol: dict, struct: str, volume_per: float, dose_1d: np.ndarray = None,
                 weight_flag: bool = True) -> float:
        """
        Get dose_1d at volume percentage

        :param sol: solution dictionary
        :param dose_1d: dose_1d in 1d
        :param struct: struct_name name for which to get the dose_1d
        :param volume_per: query the dose_1d at percentage volume
        :param weight_flag: for non uniform voxels weight flag always True
        :return: dose_1d at volume_percentage

        :Example:

        >>> Evaluation.get_dose(sol=sol, struct='PTV', volume_per=90)

        """
        x, y = Evaluation.get_dvh(sol, dose_1d=dose_1d, struct=struct, weight_flag=weight_flag)
        if np.array_equal(x, np.array([0])) and np.array_equal(y, np.array([0])):
            return 0
        f = interpolate.interp1d(100 * y, x)

        return f(volume_per)

    @staticmethod
    def get_volume(sol: dict, struct: str, dose_value_gy: float, dose_1d: np.ndarray = None,
                   weight_flag: bool = True) -> float:
        """
        Get volume at dose_1d value in Gy

        :param sol: solution dictionary
        :param dose_1d: dose_1d in 1d
        :param struct: struct_name name for which to get the dose_1d
        :param dose_value_gy: query the volume at dose_value
        :param weight_flag: for non uniform voxels weight flag always True
        :return: dose_1d at volume_percentage

        :Example:

        >>> Evaluation.get_volume(sol=sol, struct='PTV', dose_value_gy=60)

        """
        x, y = Evaluation.get_dvh(sol, dose_1d=dose_1d, struct=struct, weight_flag=weight_flag)
        if np.array_equal(x, np.array([0])) and np.array_equal(y, np.array([0])):
            return 0
        x1, indices = np.unique(x, return_index=True)
        y1 = y[indices]
        f = interpolate.interp1d(x1, 100 * y1)
        if dose_value_gy > max(x1):
            print('Warning: dose_1d value {} is greater than max dose_1d for {}'.format(dose_value_gy, struct))
            return 0
        else:
            return f(dose_value_gy)

    @staticmethod
    def get_dvh(sol: dict, struct: str, dose_1d: np.ndarray = None, weight_flag: bool = True):
        """
        Get dvh for the struct_name

        :param sol: optimal solution dictionary
        :param dose_1d: dose_1d which is not in solution dictionary
        :param struct: struct_name name
        :param weight_flag: for non uniform voxels weight flag always True
        :return: x, y --> dvh for the struct_name

        :Example:

        >>> Evaluation.get_dvh(sol=sol, struct='PTV')

        """
        inf_matrix = sol['inf_matrix']
        vox = inf_matrix.get_opt_voxels_idx(struct)
        if len(vox) == 0:
            return np.array([0]), np.array([0])  # bug fix. if single 0 it can throw error while doing interpolation
        if dose_1d is None:
            dose_1d = sol['dose_1d']
        org_sort_dose = np.sort(dose_1d[vox])
        sort_ind = np.argsort(dose_1d[vox])
        org_sort_dose = np.append(org_sort_dose, org_sort_dose[-1] + 0.01)
        x = org_sort_dose
        if weight_flag:
            # org_points_sort_spacing = my_plan._structures.opt_voxels_dict['dose_voxel_resolution_XYZ_mm']
            # org_points_sort_volume = org_points_sort_spacing[:, 0] * org_points_sort_spacing[:,
            #                                                          1] * org_points_sort_spacing[:, 2]
            # sum_weight = np.sum(org_points_sort_volume)
            org_weights = inf_matrix.get_opt_voxels_volume_cc(struct)
            org_sort_weights = org_weights[sort_ind]
            sum_weight = np.sum(org_sort_weights)
            frac_vol = inf_matrix.get_fraction_of_vol_in_calc_box(struct)
            if frac_vol is None:
                y = [1]
            else:
                y = [frac_vol]
            for j in range(len(org_sort_weights)):
                y.append(y[-1] - (org_sort_weights[j] / sum_weight)*frac_vol)
        else:
            y = np.ones(len(vox) + 1) - np.arange(0, len(vox) + 1) / len(vox)
        y[-1] = 0
        y = np.array(y)
        return x, y

    @staticmethod
    def get_max_dose(sol: dict, struct: str, dose_1d=None) -> float:
        """
        Get maximum dose_1d for the struct_name

        :param sol: optimal solution dictionary
        :param dose_1d: dose_1d which is not in solution dictionary
        :param struct: struct_name name

        :return: maximum dose_1d for the struct_name
        """
        inf_matrix = sol['inf_matrix']
        vox = inf_matrix.get_opt_voxels_idx(struct)
        if len(vox) == 0:
            return 0
        if dose_1d is None:
            dose_1d = sol['dose_1d']
        return np.max(dose_1d[vox])

    @staticmethod
    def get_mean_dose(sol: dict, struct: str, dose_1d=None) -> np.ndarray:
        """
                Get mean dose_1d for the struct_name

                :param sol: optimal solution dictionary
                :param dose_1d: dose_1d which is not in solution dictionary
                :param struct: struct_name name

                :return: mean dose_1d for the struct_name
                """
        inf_matrix = sol['inf_matrix']
        vox = inf_matrix.get_opt_voxels_idx(struct)
        if len(vox) == 0:
            return np.array(0)
        if dose_1d is None:
            dose_1d = sol['dose_1d']
        frac_vol = inf_matrix.get_fraction_of_vol_in_calc_box(struct)
        mean_dose = (1 / sum(inf_matrix.get_opt_voxels_volume_cc(struct))) * (np.sum((np.multiply(inf_matrix.get_opt_voxels_volume_cc(struct), dose_1d[vox]))))
        return mean_dose * frac_vol

    @staticmethod
    def get_BED(my_plan: Plan, sol: dict = None, dose_per_fraction_1d: np.ndarray = None, alpha=1, beta=1) -> np.ndarray:
        """
        Get Biologically equivalent dose (BED) for the struct_name

        :param my_plan: Object of class Plan
        :param sol: optimal solution dictionary
        :param dose_per_fraction_1d: dose_1d which is not in solution dictionary

        :return: BED
        """

        if dose_per_fraction_1d is None:
            dose_per_fraction_1d = sol['inf_matrix'].A @ (sol['optimal_intensity'])
        bed_d = np.zeros_like(dose_per_fraction_1d)
        for i in range(int(my_plan.get_num_of_fractions())):
            bed_d = bed_d + (dose_per_fraction_1d + (dose_per_fraction_1d**2/(alpha/beta)))
        return bed_d

    @staticmethod
    def is_notebook() -> bool:
        try:
            from IPython import get_ipython
            shell = get_ipython().__class__.__name__
            if shell == 'ZMQInteractiveShell':
                return True  # Jupyter notebook or qtconsole
            elif shell == 'TerminalInteractiveShell':
                return False  # Terminal running IPython
            else:
                if 'google.colab' in str(get_ipython()):
                    return True
                else:
                    return False  # Other type (?)
        except NameError:
            return False  # Probably standard Python interpreter
        except:
            return False  # Probably standard Python interpreter



import numpy as np
from copy import deepcopy
from .data_explorer import DataExplorer


class Structures:
    """
    A class representing structures.

    - **Attributes** ::

        :param structures_dict: struct_name dictionary that contains information about the structures present in the patient's CT scan.
        :type structures_dict: dict
        :param opt_voxels_dict: It contains information about optimization voxels in the form of dictionary
        :type opt_voxels_dict: dict


    - **Methods** ::

        :get_volume_cc(struct):
            Get volume in cc for the struct_name
        :expand(struct_name, margin_mm, new_struct_name):
            Expand the struct_name with margin_mm in mm
        :shrink(struct_name, margin_mm, new_struct_name):
            Shrink the struct_name with margin_mm in mm
        :union(struct_1_name, struct_2_name, new_struct_name):
            Create union of two structures struct_1_name and struct_2_name
        :intersect(struct_1_name, struct_2_name, new_struct_name):
            Create intersect of two structures struct_1_name and struct_2_name


    """

    def __init__(self, data: DataExplorer) -> None:
        """

        :param data: object of DataExplorer Class
        """
        metadata = data.load_metadata()
        structures_dict = data.load_data(meta_data=metadata['structures'])
        opt_voxels_dict = data.load_data(meta_data=metadata['opt_voxels'])
        self.structures_dict = structures_dict
        self.opt_voxels_dict = opt_voxels_dict
        self.opt_voxels_dict['name'] = structures_dict['name']
        self._ct_voxel_resolution_xyz_mm = deepcopy(self.opt_voxels_dict['ct_voxel_resolution_xyz_mm'])
        self.preprocess_structures()

    def get_structures(self) -> list:
        """
        Returns all the struct_name names as list
        :return:
        """
        return self.structures_dict['name']

    def get_volume_cc(self, structure_name: str):
        """
        Get volume in cc for the struct_name

         :param structure_name: name of the struct_name in plan
         :return: volume of the struct_name
         """
        ind = self.structures_dict['name'].index(structure_name)
        return self.structures_dict['volume_cc'][ind]

    def get_fraction_of_vol_in_calc_box(self, structure_name: str):
        """
        Get fraction of volume in calc box for the struct_name

         :param structure_name: name of the struct_name in plan
         :return: volume of the struct_name
         """
        ind = self.structures_dict['name'].index(structure_name)
        return self.structures_dict['fraction_of_vol_in_calc_box'][ind]

    def get_structure_mask_3d(self, structure_name: str):
        """
        Get fraction of volume in calc box for the struct_name

         :param structure_name: name of the struct_name in plan
         :return: volume of the struct_name
         """
        ind = self.structures_dict['name'].index(structure_name)
        return self.structures_dict['structure_mask_3d'][ind]

    def union(self, struct_1_name: str, struct_2_name: str, new_struct_name: str = None, return_mask=False):
        """
        Create union of two structures struct_1_name and struct_2_name. If str1_or_str2 is not in structures dict,
        it will create new structures. If str1_or_str2 is in structure_dict, it will modify the struct_name

        :param struct_1_name: struct_name name for the 1st struct_name
        :param struct_2_name: struct_name name for the 2nd struct_name
        :param new_struct_name: struct_name name for the union of  struct_name 1 and 2
        :return: create union of the structures and save it to structures list
        """

        ind1 = self.structures_dict['name'].index(struct_1_name)
        ind2 = self.structures_dict['name'].index(struct_2_name)
        mask_3d_1 = self.structures_dict['structure_mask_3d'][ind1]
        mask_3d_2 = self.structures_dict['structure_mask_3d'][ind2]
        new_mask_3d_1 = mask_3d_1 | mask_3d_2

        if new_struct_name is not None:
            if new_struct_name not in self.structures_dict['name']:
                self.create_structure(new_struct_name=new_struct_name, mask_3d=new_mask_3d_1)
            else:
                self.modify_structure(struct_name=new_struct_name, mask_3d=new_mask_3d_1)
        if return_mask:
            return new_mask_3d_1

    def intersect(self, struct_1_name: str, struct_2_name: str, new_struct_name: str = None, return_mask=False):
        """
        Create intersection of two structures struct_1_name and struct_2_name. If str1_and_str2 is not in structures dict,
        it will create new structures. If str1_and_str2 is in structure_dict, it will modify the struct_name

        :param struct_1_name: struct_name name for the 1st struct_name
        :param struct_2_name: struct_name name for the 2nd struct_name
        :param new_struct_name: struct_name name for the intersection of  struct_name 1 and 2
        :param return_mask: return 3d mask
        :return: create intersection of the structures and save it to structures list
        """

        ind1 = self.structures_dict['name'].index(struct_1_name)
        ind2 = self.structures_dict['name'].index(struct_2_name)
        mask_3d_1 = self.structures_dict['structure_mask_3d'][ind1]
        mask_3d_2 = self.structures_dict['structure_mask_3d'][ind2]
        new_mask_3d_1 = mask_3d_1 & mask_3d_2

        if new_struct_name not in self.structures_dict['name']:
            self.create_structure(new_struct_name=new_struct_name, mask_3d=new_mask_3d_1)
        else:
            self.modify_structure(struct_name=new_struct_name, mask_3d=new_mask_3d_1)
        if return_mask:
            return new_mask_3d_1

    def subtract(self, struct_1_name: str, struct_2_name: str, new_struct_name: str = None, return_mask=False):
        """
        :param struct_1_name: struct_name name for the 1st struct_name
        :param struct_2_name: struct_name name for the 2nd struct_name
        :param new_struct_name: struct_name name for subtracting 2 from 1
        :param return_mask: return 3d mask of the structure
        :return: create structure1 - structure2 and save it to list
        """

        ind1 = self.structures_dict['name'].index(struct_1_name)
        ind2 = self.structures_dict['name'].index(struct_2_name)
        mask_3d_1 = deepcopy(self.structures_dict['structure_mask_3d'][ind1])
        mask_3d_2 = deepcopy(self.structures_dict['structure_mask_3d'][ind2])
        new_mask_3d_1 = mask_3d_1.astype(int) - mask_3d_2.astype(int)  # convert to integer before doing the operation
        new_mask_3d_1[new_mask_3d_1 < 0] = int(0)
        new_mask_3d_1.astype('uint8')
        if new_struct_name is not None:
            if new_struct_name not in self.structures_dict['name']:
                self.create_structure(new_struct_name=new_struct_name, mask_3d=new_mask_3d_1)
            else:
                self.modify_structure(struct_name=new_struct_name, mask_3d=new_mask_3d_1)
        if return_mask:
            return new_mask_3d_1

    def expand(self, struct_name: str = None, margin_mm: float = None, new_struct_name: str = None,
               return_mask: bool = False, mask_3d: np.ndarray = None):
        """

        Expand the struct_name with the given margin_mm.

        :param struct_name: struct_name name to expand
        :param margin_mm: margin_mm in mm
        :param new_struct_name: new struct_name name. if same as struct_name name,
        expand the same struct_name. else create new struct_name
        :return: expand and save the struct_name mask in structures dictionary
        """
        # from skimage import data, morphology, transform
        from scipy import ndimage
        if mask_3d is None:
            ind = self.structures_dict['name'].index(struct_name)
            mask_3d = self.structures_dict['structure_mask_3d'][ind]

        # getting kernel size for expansion or shrinking
        if margin_mm > 0:
            num_voxels = np.round(margin_mm / np.asarray(self._ct_voxel_resolution_xyz_mm)).astype(int)

            kernel = ndimage.generate_binary_structure(rank=3, connectivity=2)  # ball kernel
            # creating different iterations along z and xy directions
            num_iterations = int(num_voxels[0] / num_voxels[2])
            iterations_in_one_step = int(np.round(num_voxels[0] / num_iterations))
            # margin_mask_3d = ndimage.binary_dilation(mask_3d, struct_name=struct).astype(mask_3d.dtype)
            for i in range(num_iterations):
                if i == 0:
                    margin_mask_3d = ndimage.binary_dilation(mask_3d, structure=kernel,
                                                             iterations=iterations_in_one_step).astype(mask_3d.dtype)
                else:
                    flat = np.copy(kernel)
                    flat[0, :, :] = 0
                    flat[-1, :, :] = 0
                    margin_mask_3d = ndimage.binary_dilation(margin_mask_3d, structure=flat,
                                                             iterations=iterations_in_one_step).astype(mask_3d.dtype)
        elif margin_mm == 0:
            margin_mask_3d = mask_3d
        else:
            raise ValueError('Invalid margin {}'.format(margin_mm))
        if new_struct_name is not None:
            if new_struct_name not in self.structures_dict['name']:
                self.create_structure(new_struct_name=new_struct_name, mask_3d=margin_mask_3d)
            else:
                self.modify_structure(struct_name=struct_name, mask_3d=margin_mask_3d)
        if return_mask:
            return margin_mask_3d

    def shrink(self, struct_name: str = None, margin_mm: float = None, new_struct_name: str = None,
               return_mask: bool = False, mask_3d: np.ndarray = None):
        """
        Shrink the struct_name with margin_mm.

            :param struct_name: struct_name name to shrink
            :param margin_mm: margin_mm in mm
            :param new_struct_name: new struct_name name. if same as struct_name, shrink the same struct_name. else create new struct_name
            :return: shrink and save the struct_name in structures_dictionary

        """

        from scipy import ndimage
        if mask_3d is None:
            ind = self.structures_dict['name'].index(struct_name)
            mask_3d = self.structures_dict['structure_mask_3d'][ind]


        # getting kernel size for expansion or shrinking
        num_voxels = np.round(margin_mm / np.asarray(self._ct_voxel_resolution_xyz_mm)).astype(int)
        struct = ndimage.generate_binary_structure(rank=3, connectivity=2)  # ball kernel
        # creating different iterations along z and xy directions
        num_iterations = int(num_voxels[0] / num_voxels[2])
        iterations_in_one_step = int(np.round(num_voxels[0] / num_iterations))
        for i in range(num_iterations):
            if i == 0:
                margin_mask_3d = ndimage.binary_erosion(mask_3d, structure=struct,
                                                        iterations=iterations_in_one_step).astype(mask_3d.dtype)
            else:
                flat = np.copy(struct)
                flat[0, :, :] = 0
                flat[-1, :, :] = 0
                margin_mask_3d = ndimage.binary_erosion(margin_mask_3d, structure=flat,
                                                        iterations=iterations_in_one_step).astype(mask_3d.dtype)

            # margin_mask_3d = ndimage.binary_dilation(mask_3d, struct_name=struct_name).astype(mask_3d.dtype)
        if new_struct_name is not None:
            if new_struct_name not in self.structures_dict['name']:
                self.create_structure(new_struct_name=new_struct_name, mask_3d=margin_mask_3d)
            else:
                self.modify_structure(struct_name=struct_name, mask_3d=margin_mask_3d)
        if return_mask:
            return margin_mask_3d

    def create_opt_structures(self, opt_params=None, clinical_criteria=None):
        # create rinds for optimization

        obj_funcs = []
        opt_params_constraints = []
        criteria = []
        if opt_params is not None:
            obj_funcs = opt_params['objective_functions'] if 'objective_functions' in opt_params else []
            opt_params_constraints = opt_params['constraints'] if 'constraints' in opt_params else []

        if clinical_criteria is not None:
            criteria = clinical_criteria.clinical_criteria_dict['criteria']

        opt_obj_cons = obj_funcs + opt_params_constraints
        print('creating rinds.. This step may take some time due to dilation')
        for ind, obj in enumerate(opt_obj_cons):
            if 'structure_def' in obj:
                if obj['structure_name'] not in self.get_structures():
                    structure_def = obj['structure_def']
                    mask_3d = self.evaluate_expression(structure_def)
                    result = mask_3d & self.get_structure_mask_3d('BODY')
                    self.create_structure(new_struct_name=obj['structure_name'], mask_3d=result)

        for ind, criterion in enumerate(criteria):
            if 'structure_def' in criterion['parameters']:
                param = criterion['parameters']
                structure_def = param['structure_def']
                if param['structure_name'] not in self.get_structures():
                    mask_3d = self.evaluate_expression(structure_def)
                    result = mask_3d & self.get_structure_mask_3d('BODY')
                    self.create_structure(new_struct_name=param['structure_name'], mask_3d=result)

        print('Optimization structures created!!')
        # for param in rind_params:
        #     self.set_opt_voxel_idx(struct_name=param['name'])
        self.preprocess_structures()

    def preprocess_structures(self):
        """
        preprocess structures to create optimization voxel indices for the struct_name
        :return:
        """
        self.opt_voxels_dict['voxel_idx'] = [None] * len(self.structures_dict['name'])
        self.opt_voxels_dict['voxel_volume_cc'] = [None] * len(self.structures_dict['name'])
        for i in range(len(self.structures_dict['name'])):
            vox_3d = self.structures_dict['structure_mask_3d'][i] * self.opt_voxels_dict['ct_to_dose_voxel_map'][0]
            # my_plan.structures_dict['voxel_idx'][i] = np.unique(vox_3d[vox_3d > 0])
            vox, counts = np.unique(vox_3d[vox_3d > 0], return_counts=True)
            self.opt_voxels_dict['voxel_idx'][i] = vox
            self.opt_voxels_dict['voxel_volume_cc'][i] = counts * np.prod(
                self._ct_voxel_resolution_xyz_mm)/1000  # calculate weight for each voxel
            # dividing by 1000 due to conversion from mm3 to cm3
            # self.opt_voxels_dict['voxel_volume_cc'][i] = counts / np.max(counts)  # calculate weight for each voxel

    def create_structure(self, new_struct_name: str, mask_3d: np.ndarray) -> None:
        """
        Create a new struct_name and append its mask to the structures_dict

        :param new_struct_name: name of the new struct_name
        :param mask_3d: 3d mask for the struct_name
        :return: create new_struct_name
        """
        for key in self.structures_dict.keys():
            if key == 'name':
                self.structures_dict['name'].append(new_struct_name)
            elif key == 'structure_mask_3d':
                self.structures_dict['structure_mask_3d'].append(mask_3d)
            elif key == 'volume_cc':
                counts = np.count_nonzero(mask_3d)
                self.structures_dict['volume_cc'].append(counts * np.prod(self._ct_voxel_resolution_xyz_mm) / 1000)
            elif key == 'fraction_of_vol_in_calc_box':  # implement it..
                counts = np.count_nonzero(mask_3d)
                volume_cc = counts * np.prod(self._ct_voxel_resolution_xyz_mm) / 1000
                ct_to_dose_map = self.opt_voxels_dict['ct_to_dose_voxel_map'][0]
                dose_mask = ct_to_dose_map >= 0
                dose_mask = dose_mask.astype('uint8')
                frac_of_mask_in_calc_box = mask_3d & dose_mask
                counts = np.count_nonzero(frac_of_mask_in_calc_box)
                volume_cc_in_calc_box = counts * np.prod(self._ct_voxel_resolution_xyz_mm) / 1000
                self.structures_dict['fraction_of_vol_in_calc_box'].append(volume_cc_in_calc_box/volume_cc)
            else:
                self.structures_dict[key].append(None)

    def modify_structure(self, struct_name: str, mask_3d: np.ndarray) -> None:
        """
        :param struct_name: name of the struct_name to be modified
        :param mask_3d: 3d mask for the struct_name
        :return: modify struct_name
        """
        ind = self.structures_dict['name'].index(struct_name)
        for key in self.structures_dict.keys():
            if key == 'name':
                self.structures_dict['name'][ind] = struct_name
            elif key == 'structure_mask_3d':
                self.structures_dict['structure_mask_3d'][ind] = mask_3d
            elif key == 'volume_cc':
                counts = np.count_nonzero(mask_3d)
                self.structures_dict['volume_cc'][ind] = (counts * np.prod(self._ct_voxel_resolution_xyz_mm) / 1000)
            elif key == 'fraction_of_vol_in_calc_box':  # implement it..
                counts = np.count_nonzero(mask_3d)
                volume_cc = counts * np.prod(self._ct_voxel_resolution_xyz_mm) / 1000
                ct_to_dose_map = self.opt_voxels_dict['ct_to_dose_voxel_map'][0]
                dose_mask = ct_to_dose_map >= 0
                dose_mask = dose_mask.astype('uint8')
                frac_of_mask_in_calc_box = mask_3d & dose_mask
                counts = np.count_nonzero(frac_of_mask_in_calc_box)
                volume_cc_in_calc_box = counts * np.prod(self._ct_voxel_resolution_xyz_mm) / 1000
                self.structures_dict['fraction_of_vol_in_calc_box'][ind] = volume_cc_in_calc_box/volume_cc
            else:
                self.structures_dict[key][ind] = None

    def delete_structure(self, structure: str):
        """
        :param structure: struct_name to be removed
        :return:
        """
        ind = self.structures_dict['name'].index(structure)
        for key in self.structures_dict.keys():
            del self.structures_dict[key][ind]

    def set_opt_voxel_idx(self, struct_name):
        ind = self.structures_dict['name'].index(struct_name)
        vox_3d = self.structures_dict['structure_mask_3d'][ind] * \
                 self.opt_voxels_dict['ct_to_dose_voxel_map'][0]
        # my_plan.structures_dict['voxel_idx'][i] = np.unique(vox_3d[vox_3d > 0])
        vox, counts = np.unique(vox_3d[vox_3d > 0], return_counts=True)
        self.opt_voxels_dict['voxel_idx'].append(vox)
        # self.opt_voxels_dict['voxel_volume_cc'].append(counts / np.max(counts))  # calculate weight for each voxel
        self.opt_voxels_dict['voxel_volume_cc'].append(
            counts * np.prod(self._ct_voxel_resolution_xyz_mm))  # calculate weight for each voxel
        self.opt_voxels_dict['name'].append(struct_name)

    def apply_operator(self, operators_stack, values_stack):
        operator = operators_stack.pop()
        right_operand = values_stack.pop()
        if isinstance(right_operand, str):
            if right_operand == "inf":
                right_operand = "500"
        left_operand = values_stack.pop()

        if isinstance(left_operand, str):
            if not left_operand.isnumeric():
                left_operand = self.get_structure_mask_3d(left_operand)
        if isinstance(right_operand, str):
            if not right_operand.isnumeric():
                right_operand = self.get_structure_mask_3d(right_operand)

        if operator == '+':
            if isinstance(right_operand, str):
                result = self.expand(mask_3d=left_operand, margin_mm=float(right_operand), return_mask=True)
            else:
                result = left_operand | right_operand
        elif operator == '-':
            if isinstance(right_operand, str):
                result = self.shrink(mask_3d=left_operand, margin_mm=float(right_operand), return_mask=True)
            else:
                result = left_operand.astype(int) - right_operand.astype(int)
                result[result < 0] = int(0)
                result.astype('uint8')

        elif operator == '|':
            result = left_operand | right_operand
        elif operator == '&':
            result = left_operand & right_operand

        values_stack.append(result)

    def evaluate_expression(self, expression):

        operators_stack = []
        values_stack = []

        i = 0
        while i < len(expression):
            if expression[i] == ' ':
                i += 1
                continue
            elif expression[i].isalpha() or expression[i].isdigit():
                j = i
                while j < len(expression) and (expression[j].isalpha() or expression[j] in '_.$^' or expression[j].isdigit()):
                    j += 1
                struct = expression[i:j]
                values_stack.append(struct)
                i = j - 1
            elif expression[i] in '+-|&':
                while operators_stack and operators_stack[-1] in '+-|&':
                    self.apply_operator(operators_stack, values_stack)
                operators_stack.append(expression[i])
            elif expression[i] == '(':
                operators_stack.append(expression[i])
            elif expression[i] == ')':
                while operators_stack[-1] != '(':
                    self.apply_operator(operators_stack, values_stack)
                operators_stack.pop()

            i += 1

        while operators_stack:
            self.apply_operator(operators_stack, values_stack)

        return values_stack[0]



import numpy as np
from shapely.geometry import LinearRing, Polygon, Point
from scipy import sparse
from copy import deepcopy
from scipy.sparse import csr_matrix
import itertools
from patchify import patchify
from typing import List, Union
from .ct import CT
from .beam import Beams
from .structures import Structures


class InfluenceMatrix:
    """
    class of influence matrix

    - **Attributes** ::

        :param beamlet_width_mm: The width of each beamlet in millimeters. Default is 2.5. Must be a multiple of 2.5.
        :param beamlet_height_mm: The height of each beamlet in millimeters. Default is 2.5. Must be a multiple of 2.5.
        :param opt_vox_xyz_res_mm: A list of integers representing the resolution for optimization voxels in the x, y, and z axes. When set to None, the original optimization voxel resolution will be used.
        :param target_structure: The target struct_name for creating the beamlets in the influence matrix. Default is 'PTV'.
        :param beamlets_dict: A dictionary containing the information about beamlets used in the plan.
        :param opt_beamlets_PTV_margin_mm: A float value representing the margin_mm around the PTV that was used in creating beamlets
        :param A: influence matrix that is generated by the get_influence_matrix() method.
        :param opt_voxels_dict: A dictionary containing the information about optimization voxels

    - **Methods** ::
        :dose_1d_to_3d(sol)
            Convert dose_1d from 1d to 3d and return dose_1d in 3d
        :dose_3d_to_1d(dose_3d)
            Convert dose_1d from 3d to 1d voxels and return dose_1d in 1d
        :fluence_2d_to_1d(fluence_2d)
            Create vector of intensities from 2d fluence maps
        :fluence_1d_to_2d(fluence_1d)
            From vector of intensities create 2d fluence maps

    """

    def __init__(self, ct: CT, structs: Structures, beams: Beams,
                 beamlet_width_mm: float = None, beamlet_height_mm: float = None, opt_vox_xyz_res_mm: List[float] = None,
                 is_full: bool = False, target_structure: str = 'PTV', opt_beamlets_PTV_margin_mm: float = 3) -> None:
        """
        Create a influence matrix object for Influence Matrix class based upon beamlet resolution and opt_vox_xyz_res_mm

        :param ct: object of class CT
        :param structs: object of class Structures
        :param beams: Object of class Beams
        :param beamlet_width_mm: beamlet width in mm. It should be multiple of 2.5, defaults to 2.5
        :param beamlet_height_mm: beamlet height in mm. It should be multiple of 2.5, defaults to 2.5
        :param target_structure: target struct_name for creating BEV beamlets, defaults to 'PTV'
        :param opt_vox_xyz_res_mm: It down-samples optimization voxels as factor of ct resolution
                e.g. opt_vox_xyz_res = [5*ct.res.x,5*ct.res.y,1*ct.res.z]. It will down-sample optimization voxels with 5 * ct res. in x direction, 5 * ct res. in y direction and 1*ct res. in z direction.
                defaults to None. When None it will use the original optimization voxel resolution.
        :param is_full: Load full or sparse matrix. defaults to False. If True, will load full matrix

        """
        if beamlet_width_mm is None and beamlet_height_mm is None:
            self.beamlet_width_mm = beams.get_beamlet_width()
            self.beamlet_height_mm = beams.get_beamlet_height()
        else:
            if beamlet_width_mm % 2.5 == 0 and beamlet_height_mm % 2.5 == 0:
                self.beamlet_width_mm = beamlet_width_mm
                self.beamlet_height_mm = beamlet_height_mm
            else:
                raise ValueError('beamlet_width_mm and beamlet_height_mm should be multiple of 2.5')
        self._beams = beams
        self._structs = structs
        self._ct = ct

        down_sample_xyz = None  # Temporary variable to check if we want to down sample or not
        if opt_vox_xyz_res_mm is not None:
            down_sample_xyz = [round(i / j) for i, j in zip(opt_vox_xyz_res_mm, ct.get_ct_res_xyz_mm())]
            if np.all(np.array(down_sample_xyz) == 1):  # if all are 1 then no down sample
                down_sample_xyz = None

        self._down_sample_xyz = down_sample_xyz
        self.is_full = is_full

        # create deepcopy of the object or else it will modify the structs object
        if hasattr(structs, 'opt_voxels_dict'):
            self.opt_voxels_dict = deepcopy(structs.opt_voxels_dict)
            # del structs.opt_voxels_dict  # remove opt_voxels_dict from structures
        else:
            self.opt_voxels_dict = deepcopy(self.opt_voxels_dict)

        # creating deepcopy so that it doesnt modify beams object
        if 'beamlets' in beams.beams_dict:
            self.beamlets_dict = deepcopy(beams.beams_dict['beamlets'])
        else:
            self.beamlets_dict = deepcopy(self.beamlets_dict['beamlets'])

        for i in range(len(beams.beams_dict['ID'])):
            self.beamlets_dict[i]['beam_id'] = beams.beams_dict['ID'][i]  # save beam_id in beamlet_dict

        self.opt_beamlets_PTV_margin_mm = opt_beamlets_PTV_margin_mm
        self.opt_voxels_dict['ct_origin_xyz_mm'] = ct.ct_dict['origin_xyz_mm']  # store ct origin from plan
        print('Creating BEV..')
        self.preprocess_beams(structure=target_structure)
        if self._down_sample_xyz is not None:
            self.pre_process_voxels()  # create new optimization voxel indices based on down-sample resolution

        if not self.is_full:
            print('Loading sparse influence matrix...')
            self.A = self.get_influence_matrix()  # create sparse influence matrix
            self.sparse_tol = float(beams.beams_dict['influenceMatrixSparse_tol'][0])
        else:
            print('Loading full influence matrix..')
            self.A = self.get_influence_matrix(self.is_full)  # create full matrix
        self._vox_map = None
        self._vox_weights = None
        print('Done')

    def get_voxel_info(self, row_number):
        row_dict = {}
        if row_number <= self.A.shape[0]:
            for i in range(len(self.opt_voxels_dict['name'])):
                if row_number in list(self.opt_voxels_dict['voxel_idx'][i]):
                    row_dict.setdefault('structures', []).append(self.opt_voxels_dict['name'][i])

            patch = np.where(
                self.opt_voxels_dict['ct_to_dose_voxel_map'][0] == row_number)  # get dose voxel index patch

            # get center of patch. It should be equivalent dose voxel center
            z_ind = (min(patch[0]) + max(patch[0])) / 2
            y_ind = (min(patch[1]) + max(patch[1])) / 2
            x_ind = (min(patch[2]) + max(patch[2])) / 2
            ct_res = self.opt_voxels_dict['ct_voxel_resolution_xyz_mm']
            ct_orig = self.opt_voxels_dict['ct_origin_xyz_mm']
            xyz_mm = [ct_orig[0] + x_ind * ct_res[0], ct_orig[1] + y_ind * ct_res[1], ct_orig[2] + z_ind * ct_res[2]]
            row_dict['position_xyz_mm'] = xyz_mm
        else:
            raise ValueError('invalid row number {}'.format(row_number))
        return row_dict

    def get_beamlet_info(self, col_number: int) -> dict:
        """

        :param col_number: col number/beamlet number in influence matrix
        :return: dictionary containing information about the beamlet
        """
        col_dict = {}
        for ind in range(len(self.beamlets_dict)):
            if col_number in range(self.beamlets_dict[ind]['start_beamlet_idx'],
                                   self.beamlets_dict[ind]['end_beamlet_idx'] + 1):
                if ind > 0:
                    prev_beam_beamlet = self.beamlets_dict[ind - 1]['end_beamlet_idx'] + 1
                else:
                    prev_beam_beamlet = 0
                col_dict['beam_id'] = self.beamlets_dict[ind]['beam_id']
                col_dict['position_x_mm'] = self.beamlets_dict[ind]['position_x_mm'][0][col_number - prev_beam_beamlet]
                col_dict['position_y_mm'] = self.beamlets_dict[ind]['position_y_mm'][0][col_number - prev_beam_beamlet]
                col_dict['width_mm'] = self.beamlets_dict[ind]['width_mm'][0][col_number - prev_beam_beamlet]
                col_dict['height_mm'] = self.beamlets_dict[ind]['height_mm'][0][col_number - prev_beam_beamlet]
        return col_dict

    def dose_1d_to_3d(self, sol: dict = None, dose_1d: np.array = None) -> np.ndarray:
        """
        Create 3d dose_1d from dose_1d in 1d voxels. 3d array have same resolution as CT

        :param sol: solution dictionary from optimization
        :param dose_1d: dose_1d in 1d. Optional
        :return: dose_1d in 3d


        """
        dose_vox_map = self.opt_voxels_dict['ct_to_dose_voxel_map'][0]
        # dose_1d = my_plan.opt_voxels_dict['dose_1d']
        if dose_1d is None:
            if 'dose_1d' not in sol:
                dose_1d = sol['inf_matrix'].A * sol['optimal_intensity']  # multiply it with num fractions
            else:
                dose_1d = sol['dose_1d']
        dose_3d = np.zeros_like(dose_vox_map, dtype='float32')
        inds = np.unique(dose_vox_map[dose_vox_map >= 0])
        a = np.where(np.isin(dose_vox_map, inds))
        dose_3d[a] = dose_1d[dose_vox_map[a]]
        return dose_3d

    def dose_3d_to_1d(self, dose_3d: np.ndarray) -> np.array:
        """
        Get dose_1d in 1d voxels for the given influence matrix from 3d dose_1d

        :param dose_3d: 3d dose_1d
        :return: dose_1d in 1d voxel indices
        """
        dose_vox_map = self.opt_voxels_dict['ct_to_dose_voxel_map'][0]
        inds = np.unique(dose_vox_map[dose_vox_map >= 0])
        dose_1d = np.zeros_like(inds, dtype=float)
        a = np.where(np.isin(dose_vox_map, inds))
        dose_1d[dose_vox_map[a]] = dose_3d[a]
        return dose_1d

    def fluence_2d_to_1d(self, fluence_2d: List[np.ndarray]) -> np.array:
        """
        Create vector of intensities from 2d fluence maps

        :param fluence_2d: 2d fluence as list of nd array with same length as number of beams_dict
        :return: fluence in 1d for beamlet indices
        """
        fluence_1d = np.zeros((self.A.shape[1]))
        for ind in range(len(self.beamlets_dict)):
            maps = self.beamlets_dict[ind]['beamlet_idx_2d_finest_grid']
            numRows = np.size(maps, 0)
            numCols = np.size(maps, 1)
            # wMaps = np.zeros((numRows, numCols))
            for r in range(numRows):
                for c in range(numCols):
                    if maps[r, c] >= 0:
                        curr = maps[r, c]
                        fluence_1d[curr] = fluence_2d[ind][r, c]
        return fluence_1d

    def fluence_1d_to_2d(self, fluence_1d: np.array = None, sol: dict = None) -> List[np.ndarray]:
        """
        Create 2d fluence maps from vector of intensities.

        :param fluence_1d: 2d fluence as vector
        :param sol: solution from optimization
        :return: 2d fluence as a list for each beam

        """
        if fluence_1d is None:
            fluence_1d = sol['optimal_intensity']

        wMaps = []
        for ind in range(len(self.beamlets_dict)):
            maps = self.beamlets_dict[ind]['beamlet_idx_2d_finest_grid']
            numRows = np.size(maps, 0)
            numCols = np.size(maps, 1)
            wMaps.append(np.zeros((numRows, numCols)))
            # wMaps = np.zeros((numRows, numCols))
            for r in range(numRows):
                for c in range(numCols):
                    if maps[r, c] >= 0:
                        curr = maps[r, c]
                        wMaps[ind][r, c] = fluence_1d[curr]
                        # wMaps[r, c] = optimal_intensity[curr]

        return wMaps

    @staticmethod
    def sol_change_inf_matrix(sol: dict, inf_matrix) -> dict:
        """
        Create a new solution by changing the basis of current solution.
        It will create a solution with same number of beamlets and voxels as inf_matrix


        :param sol: solution for which influence matrix is changed
        :param inf_matrix: object of class Influence matrix
        :return: new solution dictionary having same number of beamlets and voxels as inf_matrix
        """
        new_sol = dict()
        if sol['optimal_intensity'].shape[0] < inf_matrix.A.shape[1]:
            optimal_intensity = sol['inf_matrix'].fluence_1d_to_2d(fluence_1d=sol['optimal_intensity'])
            new_sol['optimal_intensity'] = inf_matrix.fluence_2d_to_1d(optimal_intensity)
        elif sol['optimal_intensity'].shape[0] == inf_matrix.A.shape[1]:
            new_sol['optimal_intensity'] = sol['optimal_intensity']
        else:
            raise ValueError("Beamlet resolution should be greater than or equal to beamlets for inf_matrix")

        # new_sol['dose_1d'] = inf_matrix.A * new_sol['optimal_intensity']

        # dose_3d = sol['inf_matrix'].dose_1d_to_3d(dose_1d=sol['dose_1d'])
        # new_sol['dose_1d'] = inf_matrix.dose_3d_to_1d(dose_3d=dose_3d)

        new_sol['inf_matrix'] = inf_matrix
        return new_sol

    def create_down_sample(self, beamlet_width_mm: float = None, beamlet_height_mm: float = None,
                           opt_vox_xyz_res_mm: List[float] = None,
                           overwrite: bool = False, remove_corner_beamlets: bool = False):
        if overwrite:
            new_inf_matrix = self
        else:
            new_inf_matrix = deepcopy(self)
        if beamlet_width_mm is None:
            beamlet_width_mm = self.beamlet_width_mm
            beamlet_height_mm = self.beamlet_height_mm
        new_inf_matrix.beamlet_width_mm = beamlet_width_mm
        new_inf_matrix.beamlet_height_mm = beamlet_height_mm
        new_inf_matrix.beamlets_dict = deepcopy(self._beams.beams_dict['beamlets'])
        for i in range(len(self._beams.beams_dict['ID'])):
            new_inf_matrix.beamlets_dict[i]['beam_id'] = deepcopy(self._beams.beams_dict['ID'][i])  # save beam_id in beamlet_dict

        for i in range(len(new_inf_matrix.beamlets_dict)):
            new_inf_matrix.beamlets_dict[i]['beamlet_idx_2d_finest_grid'] = deepcopy(self.beamlets_dict[i]['beamlet_idx_2d_finest_grid'])
        down_sample_xyz = None  # Temporary variable to check if we want to down sample or not
        if opt_vox_xyz_res_mm is not None:
            down_sample_xyz = [round(i / j) for i, j in zip(opt_vox_xyz_res_mm, self.opt_voxels_dict['ct_voxel_resolution_xyz_mm'])]
            if np.all(np.array(down_sample_xyz) == 1):  # if all are 1 then no down sample
                down_sample_xyz = None
        new_inf_matrix._down_sample_xyz = down_sample_xyz
        new_inf_matrix.preprocess_beams(remove_corner_beamlets=remove_corner_beamlets)
        new_inf_matrix.pre_process_voxels()
        A = new_inf_matrix.get_influence_matrix(is_full=new_inf_matrix.is_full)
        new_inf_matrix.A = A
        return new_inf_matrix

    def get_influence_matrix(self, is_full=False):
        """

        Load influence matrix based on the beamlets and voxels.

        :param beams: object of class Beams
        :param is_full: get full or sparse matrix. Default to True.
        :return: full or sparse matrix
        """
        data_beamlet_width = self._beams.get_beamlet_width()
        data_beamlet_height = self._beams.get_beamlet_height()

        if not is_full:
            if self.beamlet_width_mm > data_beamlet_width or \
                    self.beamlet_height_mm > data_beamlet_height or self._down_sample_xyz is not None:
                inf_matrix_sparse = self.A
            else:
                # deepcopy so that it doesnt modify
                inf_matrix_sparse = deepcopy(self._beams.beams_dict['influenceMatrixSparse'])
                # if 'influenceMatrixSparse' in self._beams.beams_dict:
                del self._beams.beams_dict['influenceMatrixSparse']

            for ind in range(len(self.beamlets_dict)):
                opt_beamlets = self.beamlets_dict[ind]['opt_beamlets_ids']

                if ind == 0:
                    if self.beamlet_width_mm > data_beamlet_width or self.beamlet_height_mm > data_beamlet_height:
                        print('parsing influence matrix for beam {}'.format(ind))
                        inf_matrix = sparse.hstack(
                            [csr_matrix(inf_matrix_sparse[:, np.unique(opt_beamlets[i])].sum(axis=1)) for i in
                             range(len(opt_beamlets))], format='csr')
                    else:
                        if self._down_sample_xyz is None:
                            inf_matrix = inf_matrix_sparse[ind][:, opt_beamlets]
                else:
                    if self.beamlet_width_mm > data_beamlet_width or self.beamlet_height_mm > data_beamlet_height:
                        print('parsing influence matrix for beam {}'.format(ind))
                        inf_matrix_2 = sparse.hstack(
                            [csr_matrix(inf_matrix_sparse[:, np.unique(opt_beamlets[i])].sum(axis=1)) for i in
                             range(len(opt_beamlets))], format='csr')
                        inf_matrix = sparse.hstack(
                            [inf_matrix, inf_matrix_2], format='csr')
                    else:
                        if self._down_sample_xyz is None:
                            inf_matrix = sparse.hstack(
                                [inf_matrix, inf_matrix_sparse[ind][:, opt_beamlets]], format='csr')
            # if del_org_matrix:
            if self._down_sample_xyz is not None:
                if self.beamlet_width_mm <= data_beamlet_width or self.beamlet_height_mm <= data_beamlet_height:
                    inf_matrix = inf_matrix_sparse
                print('creating influence matrix for down sample voxels..')
                inf_matrix = sparse.vstack(
                    [csr_matrix((sparse.diags(self._vox_weights[i]).dot(inf_matrix[self._vox_map[i], :])).sum(axis=0))
                     for i
                     in
                     range(len(self._vox_map))], format='csr')
        else:
            if self.beamlet_width_mm > data_beamlet_width or self.beamlet_height_mm > data_beamlet_height or self._down_sample_xyz is not None:
                inf_matrix_full = self.A
            else:
                inf_matrix_full = self._beams.beams_dict['influenceMatrixFull']
                del self._beams.beams_dict['influenceMatrixFull']

            for ind in range(len(self.beamlets_dict)):
                opt_beamlets = self.beamlets_dict[ind]['opt_beamlets_ids']

                if ind == 0:
                    if self.beamlet_width_mm > data_beamlet_width or self.beamlet_height_mm > data_beamlet_height:
                        print('parsing full influence matrix for beam {}'.format(ind))
                        inf_matrix = np.hstack(
                            [inf_matrix_full[:, np.unique(opt_beamlets[i])].sum(axis=1) for i in
                             range(len(opt_beamlets))])
                    else:
                        if self._down_sample_xyz is None:
                            inf_matrix = inf_matrix_full[ind][:, opt_beamlets]
                else:
                    if self.beamlet_width_mm > data_beamlet_width or self.beamlet_height_mm > data_beamlet_height:
                        print('parsing full influence matrix for beam {}'.format(ind))
                        inf_matrix_2 = np.hstack(
                            [inf_matrix_full[:, np.unique(opt_beamlets[i])].sum(axis=1) for i in
                             range(len(opt_beamlets))])
                        inf_matrix = np.hstack([inf_matrix, inf_matrix_2])
                    else:
                        if self._down_sample_xyz is None:
                            inf_matrix = np.hstack([inf_matrix, inf_matrix_full[ind][:, opt_beamlets]])

                # down sampling voxels
            if self._down_sample_xyz is not None:
                if self.beamlet_width_mm <= data_beamlet_width or self.beamlet_height_mm <= data_beamlet_height:
                    inf_matrix = inf_matrix_full
                print('creating influence matrix for down sample voxels..')
                inf_matrix = np.vstack(
                    [(np.diags(self._vox_weights[i]).dot(inf_matrix[self._vox_map[i], :])).sum(axis=0)
                     for i in range(len(self._vox_map))])

        return inf_matrix

    def create_BEV_mask_from_contours(self, ind: int, structure: str = 'PTV',
                                      margin_mm: float = None) -> np.ndarray:
        """

        Since beams object contain projection of struct_name contours on BEV, this function helps to create mask from those contours on BEV

        :param beams: object of class Beams
        :param ind: indices of the beam in beamlet dictionary
        :param structure: struct_name for which contours to be converted to mask
        :param margin_mm: expand the contours in mm. defaults to None. If it is not none, it will expand the contours
        :return:
        """
        if margin_mm is None and structure == 'PTV':
            margin_mm = self.opt_beamlets_PTV_margin_mm
        elif margin_mm is None:
            margin_mm = 0

        # get PTV contour from data
        contours = self._beams.beams_dict['BEV_structure_contour_points'][ind][structure]
        # ind = my_plan.beamlets_dict['ID'].index(beam_id)
        # contours = my_plan._beams_contours[ind][struct_name]

        # for each contour create polygon and get beamlets inside the polygon and create mask for it
        for count_num in range(len(contours)):
            polygon = []
            for j in contours[count_num]:
                polygon.append((j[0], j[1]))
            r = LinearRing(polygon)
            s = Polygon(r)
            # add margin_mm around polygon
            if margin_mm == 0:
                shapely_poly = s
            else:
                shapely_poly = Polygon(s.buffer(margin_mm), [r])
            #             poly_coordinates = np.array(list(t.exterior.coords))

            # create beam map from beamlet coordinates and mask the beamlets inside shapely polygon
            beamlets = self.beamlets_dict[ind]
            x_positions = beamlets['position_x_mm'][0]
            y_positions = beamlets['position_y_mm'][0]
            x_min_max_sort = np.sort(np.unique(x_positions))
            y_max_min_sort = np.sort(np.unique(y_positions))[::-1]
            points = []
            for i in range(len(beamlets['id'][0])):
                x_coord = beamlets['position_x_mm'][0][i]
                y_coord = beamlets['position_y_mm'][0][i]
                points.append(Point(x_coord, y_coord))
            valid_points = []
            valid_points.extend(filter(shapely_poly.contains, points))
            x_and_y = [(a.x, a.y) for a in valid_points]
            XX, YY = np.meshgrid(x_min_max_sort, y_max_min_sort)
            w_all = np.column_stack((x_positions, y_positions))
            if count_num == 0:
                mask = np.zeros_like(XX, dtype=bool)
            for row in range(XX.shape[0]):
                for col in range(XX.shape[1]):
                    ind = np.where((w_all[:, 0] == XX[row, col]) & (w_all[:, 1] == YY[row, col]))[0][0]

                    # if (beamlets[ind]['position_x_mm'], beamlets[ind]['position_y_mm']) in x_and_y:
                    if (beamlets['position_x_mm'][0][ind], beamlets['position_y_mm'][0][ind]) in x_and_y:
                        # beam_map[row, col] = beamlets[ind]['id']
                        mask[row, col] = True
        return mask

    def preprocess_beams(self, structure='PTV', remove_corner_beamlets=False):
        """
        Preprocess beams to create beamlets dictionary based on beamlet resolution.

        :param beams: object of class Beams
        :param structure: projection of the struct_name on BEV for which beamlets to be considered for creating influence matrix. defaults to PTV
        :param remove_corner_beamlets: If remove corner beamlet is true, it will remove the corner beamlets during down sampling.
        :return: beamlets for processing influence matrix
        """

        for ind in range(len(self.beamlets_dict)):
            # ind = my_plan.beamlets_dict['ID'].index(beam_id)
            mask = self.create_BEV_mask_from_contours(ind=ind, structure=structure,
                                                      margin_mm=self.opt_beamlets_PTV_margin_mm)
            beam_2d_grid = self.create_beamlet_idx_2d_finest_grid(ind=ind)
            beamlets = self.beamlets_dict[ind]

            # creating beam_map of original resolution so that mask can be multiplied with it
            beam_map = self.get_orig_res_2d_grid(ind)
            beam_map = np.multiply((beam_map + int(1)), mask)  # add and subtract one to maintain 0th beamlet
            beam_map = beam_map - int(1)  # subtract one again to get original beamlets

            # get mask and beam_map in 2.5mm resolution
            beamlet_ind = np.unique(beam_map.flatten())
            beamlet_ind = beamlet_ind[beamlet_ind >= 0]
            a = np.where(np.isin(beam_2d_grid, beamlet_ind))
            mask_2d_grid = np.zeros_like(beam_2d_grid, dtype=bool)
            mask_2d_grid[a] = True
            beam_2d_grid = (beam_2d_grid + int(1)) * mask_2d_grid  # add and subtract 1 to retain ids
            beam_2d_grid = beam_2d_grid - int(1)
            beam_map = beam_2d_grid

            # get opt beamlets for down_sample grid as list of original inf_matrix beamlet
            if self.beamlet_width_mm > self._beams.get_beamlet_width() or \
                    self.beamlet_height_mm > self._beams.get_beamlet_height():
                down_sample_2d_grid = self.down_sample_2d_grid(ind=ind, beamlet_width_mm=self.beamlet_width_mm,
                                                               beamlet_height_mm=self.beamlet_height_mm)
                down_sample_2d_grid = (down_sample_2d_grid + int(1)) * mask_2d_grid  # add and subtract 1 to retain ids
                down_sample_2d_grid = down_sample_2d_grid - int(1)
                down_sample_beamlets, counts = np.unique(np.sort(down_sample_2d_grid[down_sample_2d_grid >= 0]),
                                                         return_counts=True)
                # remove corner beamlets in coarse resolution and updating down sample map
                if remove_corner_beamlets:
                    count_ind = np.where(counts >= (self.beamlet_width_mm / 2.5) * (self.beamlet_height_mm / 2.5))
                    down_sample_beamlets = down_sample_beamlets[count_ind]
                    # updating down sample grid after removing corner beamlets
                    inds_down_sample = down_sample_2d_grid == down_sample_beamlets[:, None,
                                                              None]  # keep only down sample beamlets and remove others
                    down_sample_2d_grid[~np.any(inds_down_sample, axis=0)] = -1

                a = np.where(np.isin(down_sample_2d_grid, down_sample_beamlets))
                mask_2d_grid = np.zeros_like(beam_2d_grid, dtype=bool)
                mask_2d_grid[a] = True
                # actual_beamlets = beam_2d_grid[a]
                actual_beamlets = self.beamlets_dict[ind]['beamlet_idx_2d_finest_grid'][a]
                sampled_beamlets = down_sample_2d_grid[a]
                b = [np.where(sampled_beamlets == down_sample_beamlets[i]) for i in range(len(down_sample_beamlets))]
                opt_beamlets = [actual_beamlets[i] for i in b]

                beam_map = down_sample_2d_grid
            else:
                opt_beamlets = np.unique(np.sort(beam_map[beam_map >= 0]))

            # make beamlets continuous
            std_map = self.sort_beamlets(beam_map)
            if ind == 0:
                beam_map = std_map
            else:
                beam_map = std_map + int(
                    np.amax(self.beamlets_dict[ind - 1]['beamlet_idx_2d_finest_grid']) + 1) * mask_2d_grid
            standInd = np.unique(np.sort(beam_map.flatten()))
            self.beamlets_dict[ind]['beamlet_idx_2d_finest_grid'] = beam_map
            # my_plan.beamlets_dict.setdefault('structure_mask_2dgrid', []).append(mask_2d_grid)
            self.beamlets_dict[ind]['start_beamlet_idx'] = standInd[1]
            self.beamlets_dict[ind]['end_beamlet_idx'] = np.amax(self.beamlets_dict[ind]['beamlet_idx_2d_finest_grid'])
            self.beamlets_dict[ind]['opt_beamlets_ids'] = opt_beamlets

            # update beamlet dict
            if self.beamlet_width_mm > self._beams.get_beamlet_width() or \
                    self.beamlet_height_mm > self._beams.get_beamlet_height():
                pass
            else:
                self.beamlets_dict[ind]['position_x_mm'][0] = beamlets['position_x_mm'][0][opt_beamlets]
                self.beamlets_dict[ind]['position_y_mm'][0] = beamlets['position_y_mm'][0][opt_beamlets]
                self.beamlets_dict[ind]['width_mm'][0] = beamlets['width_mm'][0][opt_beamlets]
                self.beamlets_dict[ind]['height_mm'][0] = beamlets['height_mm'][0][opt_beamlets]
                self.beamlets_dict[ind]['MLC_leaf_idx'][0] = beamlets['MLC_leaf_idx'][0][opt_beamlets]
            del self.beamlets_dict[ind]['id']
            # my_plan.beams_dict.setdefault('opt_beamlets_ids', []).append(standInd[1:])

    def create_beamlet_idx_2d_finest_grid(self, ind: int) -> np.ndarray:
        """
        Create beamlet idx in 2d grid of 2.5mm resolution. i.e. each element in matrix is 2.5mm*2.5mm

        :param ind: indices for the beam in beamlets dictionary
        :return:
        """
        # ind = my_plan.beamlets_dict['ID'].index(beam_id)
        beamlets = self.beamlets_dict[ind]
        x_positions = beamlets['position_x_mm'][0] - beamlets['width_mm'][0] / 2
        y_positions = beamlets['position_y_mm'][0] + beamlets['height_mm'][0] / 2
        right_ind = np.argmax(x_positions)
        bottom_ind = np.argmin(y_positions)
        w_all = np.column_stack((x_positions, y_positions))  # top left corners of all beamlets
        x_coord = np.arange(np.min(x_positions), np.max(x_positions) + beamlets['width_mm'][0][right_ind], 2.5)
        y_coord = np.arange(np.max(y_positions), np.min(y_positions) - beamlets['height_mm'][0][bottom_ind], -2.5)
        XX, YY = np.meshgrid(x_coord, y_coord)
        beamlet_idx_2d_grid = np.ones_like(XX, dtype=int)
        beamlet_idx_2d_grid = beamlet_idx_2d_grid * int(-1)
        for row in range(XX.shape[0]):
            for col in range(XX.shape[1]):
                ind = np.where((w_all[:, 0] == XX[row, col]) & (w_all[:, 1] == YY[row, col]))
                if np.size(ind) > 0:
                    ind = ind[0][0]
                    num_width = int(beamlets['width_mm'][0][ind] / 2.5)
                    num_height = int(beamlets['height_mm'][0][ind] / 2.5)
                    beamlet_idx_2d_grid[row:row + num_height, col:col + num_width] = ind

        return beamlet_idx_2d_grid

    def down_sample_2d_grid(self, ind: int, beamlet_width_mm: float = 5.0,
                            beamlet_height_mm: float = 5.0) -> np.ndarray:
        """
        create down sample beam map based on beamlet width and height

        :param ind: indices for the beam in beamlets dictionary
        :param beamlet_width_mm:
        :param beamlet_height_mm:
        :return: down sampled 2d grid

        """
        # ind = my_plan.beamlets_dict['ID'].index(beam_id)
        beamlets = self.beamlets_dict[ind]
        x_positions = beamlets['position_x_mm'][0] - beamlets['width_mm'][0] / 2
        y_positions = beamlets['position_y_mm'][0] + beamlets['height_mm'][0] / 2
        right_ind = np.argmax(x_positions)
        bottom_ind = np.argmin(y_positions)
        # w_all = np.column_stack((x_positions, y_positions))  # top left corners of all beamlets
        x_coord = np.arange(np.min(x_positions), np.max(x_positions) + beamlets['width_mm'][0][right_ind], 2.5)
        y_coord = np.arange(np.max(y_positions), np.min(y_positions) - beamlets['height_mm'][0][bottom_ind], -2.5)
        XX, YY = np.meshgrid(x_coord, y_coord)
        beamlet_resample_2d_grid = np.ones_like(XX, dtype=int)
        beamlet_resample_2d_grid = beamlet_resample_2d_grid * int(-1)
        row_col_covered = []
        ind = 0
        for row in range(XX.shape[0]):
            for col in range(XX.shape[1]):
                # ind = np.where((w_all[:, 0] == XX[row, col]) & (w_all[:, 1] == YY[row, col]))
                # if np.size(ind) > 0:
                #     ind = ind[0][0]
                if [row, col] not in row_col_covered:
                    num_width = int(beamlet_width_mm / 2.5)
                    num_height = int(beamlet_height_mm / 2.5)
                    beamlet_resample_2d_grid[row:row + num_height, col:col + num_width] = ind
                    ind = ind + 1
                    rows = np.arange(row, row + num_height)
                    cols = np.arange(col, col + num_width)
                    for r in itertools.product(rows, cols):
                        row_col_covered.append([r[0], r[1]])
        return beamlet_resample_2d_grid

    def get_orig_res_2d_grid(self, ind: int) -> np.ndarray:
        """
        Create beam map 2d grid in the original resolution where each element in matrix is the size of respective beamlet.

        :param ind: indices for the beam in beamlets dictionary
        :return: beam map 2d grid in original resolution

        """
        beamlets = self.beamlets_dict[ind]
        x_positions = beamlets['position_x_mm'][0]
        y_positions = beamlets['position_y_mm'][0]
        x_min_max_sort = np.sort(np.unique(x_positions))
        y_max_min_sort = np.sort(np.unique(y_positions))[::-1]
        XX, YY = np.meshgrid(x_min_max_sort, y_max_min_sort)
        w_all = np.column_stack((x_positions, y_positions))
        beam_map = np.zeros_like(XX, dtype=int)
        for row in range(XX.shape[0]):
            for col in range(XX.shape[1]):
                b_ind = np.where((w_all[:, 0] == XX[row, col]) & (w_all[:, 1] == YY[row, col]))[0][0]
                # beam_map[row, col] = beamlets[ind]['id']+int(1)# adding one so that 0th beamlet is retained
                beam_map[row, col] = b_ind  # + int(1)  # adding one so that 0th beamlet is retained
        # beam_map = np.multiply(beam_map, mask)
        # beam_map = beam_map - int(1)  # subtract one again to get original beamlets
        return beam_map

    def get_bev_2d_grid(self, beam_id: Union[int, List[int]] = None, ind: int = None,
                        finest_grid: bool = False) -> Union[np.ndarray, List[np.ndarray]]:

        """
        get BEV in 2d grid in original resolution where beamlet index is the column number in influence matrix

        :param ind: idx of the beam in beamlets_dict.
        It can be int or List[int]. If int, ndarray is return, If List[int], list[ndarray] is returned
        :param beam_id: beam_id of the beam.
        It can be int or List[int]. If int, ndarray is return, If List[int], list[ndarray] is returned
        :param finest_grid:Default to False. If set to true, it will return 2d grid in finest resolution
        :return: ndarray/List[ndarray]
        """

        if beam_id is not None:
            ind = []
            if isinstance(beam_id, int):
                ind = [i for i in range(len(self.beamlets_dict)) if
                       self.beamlets_dict[i]['beam_id'] == beam_id]
                if not ind:
                    raise ValueError("invalid beam id {}".format(beam_id))
            elif isinstance(beam_id, list):
                for idx in beam_id:
                    try:
                        ind_1 = [i for i in range(len(self.beamlets_dict)) if
                                 self.beamlets_dict[i]['beam_id'] == idx][0]
                        ind.append(ind_1)
                    except:
                        raise ValueError("invalid beam id {}".format(idx))
        # Bug fix. in case ind is int make it as list
        if isinstance(ind, int):
            ind = [ind]
        beam_orig = []
        for b in ind:
            beam_map = self.beamlets_dict[b]['beamlet_idx_2d_finest_grid']
            if not finest_grid:
                rowsNoRepeat = [0]
                for i in range(1, np.size(beam_map, 0)):
                    if (beam_map[i, :] != beam_map[rowsNoRepeat[-1], :]).any():
                        rowsNoRepeat.append(i)
                colsNoRepeat = [0]
                for j in range(1, np.size(beam_map, 1)):
                    if (beam_map[:, j] != beam_map[:, colsNoRepeat[-1]]).any():
                        colsNoRepeat.append(j)
                beam_map = beam_map[np.ix_(np.asarray(rowsNoRepeat), np.asarray(colsNoRepeat))]

                # add this part in case remove corner beamlets create issue for getting original resolution
                remove_row = []
                remove_col = []
                for row in range(beam_map.shape[0]):
                    if row < beam_map.shape[0] - 1:
                        prev_elem = beam_map[row, :][beam_map[row, :] > -1]
                        next_elem = beam_map[row + 1, :][beam_map[row + 1, :] > -1]
                        matching_elements = np.intersect1d(prev_elem, next_elem)
                        if len(matching_elements) >= 1:
                            if len(prev_elem) <= len(next_elem):
                                remove_row.append(row)
                            else:
                                remove_row.append(row + 1)
                for col in range(beam_map.shape[1]):
                    if col < beam_map.shape[1] - 1:
                        prev_elem = beam_map[:, col][beam_map[:, col] > -1]
                        next_elem = beam_map[:, col + 1][beam_map[:, col + 1] > -1]
                        matching_elements = np.intersect1d(prev_elem, next_elem)
                        if len(matching_elements) >= 1:
                            if len(prev_elem) <= len(next_elem):
                                remove_col.append(col)
                            else:
                                remove_col.append(col + 1)
                mask_rows = np.ones(beam_map.shape[0], dtype=bool)
                mask_columns = np.ones(beam_map.shape[1], dtype=bool)
                if len(remove_row) > 0:
                    mask_rows[remove_row] = False
                if len(remove_col) > 0:
                    mask_columns[remove_col] = False
                beam_map = beam_map[mask_rows][:, mask_columns]

            beam_orig.append(beam_map)
        if len(beam_orig) == 1:
            beam_orig = beam_orig[0]  # return ndarray in case if it is not list

        return beam_orig

    def down_sample_voxels(self, down_sample_xyz: List[int] = None):
        """
        This method creates new ct to dose_1d voxel map array based on down sampled voxels.
        It also outputs map between down sampled voxel indices to original voxel indices for down sampling influence matrix
        :param down_sample_xyz: It down-samples optimization voxels as factor of ct resolution
                e.g. _down_sample_xyz = [5,5,1]. It will down-sample optimization voxels with 5 * ct res. in x direction, 5 * ct res. in y direction and 1*ct res. in z direction.
                defaults to None. When None it will use the original optimization voxel resolution.
        :return: voxel map(list containing map between down-sampled indices to original indices), voxel weight (weighted average of voxels in down sample
         and ct to dose_1d voxel map
        """

        vox_3d = self.opt_voxels_dict['ct_to_dose_voxel_map'][0]
        # dose_to_ct_int = np.round(np.array(self.opt_voxels_dict['dose_voxel_resolution_xyz_mm']) /
        #                           np.array(self.opt_voxels_dict['ct_voxel_resolution_xyz_mm']))
        # dose_to_ct_int = dose_to_ct_int.astype(int)

        # using patchify
        use_patchify = True
        if use_patchify:
            print('reindexing voxels...')
            vox_map = []
            vox_weights = []
            slices, height, width = down_sample_xyz[2], down_sample_xyz[1], down_sample_xyz[0]
            patches = patchify(vox_3d, (slices, height, width), step=(slices, height, width))
            # pat = np.vstack(patches)
            # pat = np.vstack(pat)
            pat = patches.reshape(-1, slices, height, width)
            count = 0
            for i in range(np.shape(pat)[0]):
                if np.any(pat[i] > -1):
                    vox_inds, weights = np.unique(np.sort(pat[i][pat[i] >= 0]), return_counts=True)
                    weight = weights / np.sum(weights)  # calculate weight for each voxel
                    # vox_map[vox_inds] = np.column_stack((vox_inds, count * np.ones_like(vox_inds), weight))
                    vox_map.append(vox_inds)
                    vox_weights.append(weight)
                    # pat[i] = np.ones((1, 5, 5), dtype=int)*int(count)
                    pat[i][pat[i] > -1] = int(count)
                    count = count + 1
            unfold_shape = patches.shape
            output_c = unfold_shape[0] * unfold_shape[3]
            output_h = unfold_shape[1] * unfold_shape[4]
            output_w = unfold_shape[2] * unfold_shape[5]
            pat_reshape = pat.reshape(patches.shape)
            down_sample_3d = pat_reshape.transpose([0, 3, 1, 4, 2, 5]).reshape(output_c, output_h, output_w)
            if down_sample_3d.shape != vox_3d.shape:
                down_sample_3d = np.pad(
                    down_sample_3d,
                    [(0, vox_3d.shape[i] - down_sample_3d.shape[i]) for i in range(len(down_sample_3d.shape))],
                    "constant", constant_values=-1)

        # using pytorch
        use_torch = False
        if use_torch:
            import torch
            vox_map = []
            vox_weights = []
            vox_3d_torch = torch.from_numpy(vox_3d)
            kc, kh, kw = down_sample_xyz[2], down_sample_xyz[1], down_sample_xyz[0]  # xyz kernel size
            dc, dh, dw = down_sample_xyz[2], down_sample_xyz[1], down_sample_xyz[0]  # xyz stride
            patches = vox_3d_torch.unfold(0, kc, dc).unfold(1, kh, dh).unfold(2, kw, dw)
            unfold_shape = patches.size()
            patches = patches.contiguous().view(-1, kc, kh, kw)
            count = 0
            for pat in patches:
                if torch.any(pat > -1):
                    elem, ind = torch.sort(pat[pat >= 0])
                    vox_inds, weights = torch.unique(elem, return_counts=True)
                    weight = weights / torch.sum(weights)  # calculate weight for each voxel
                    # vox_map[vox_inds] = np.column_stack((vox_inds, count * np.ones_like(vox_inds), weight))
                    vox_map.append(vox_inds.numpy())
                    vox_weights.append(weight.numpy())
                    # pat[i] = np.ones((1, 5, 5), dtype=int)*int(count)
                    pat[pat > -1] = count
                    count = count + 1
            # Reshape back
            patches_orig = patches.view(unfold_shape)
            output_c = unfold_shape[0] * unfold_shape[3]
            output_h = unfold_shape[1] * unfold_shape[4]
            output_w = unfold_shape[2] * unfold_shape[5]
            patches_orig = patches_orig.permute(0, 3, 1, 4, 2, 5).contiguous()
            down_sample_3d_torch = patches_orig.view(output_c, output_h, output_w)
            down_sample_3d = down_sample_3d_torch.numpy()
            if down_sample_3d.shape != vox_3d.shape:
                down_sample_3d = np.pad(
                    down_sample_3d,
                    [(0, vox_3d.shape[i] - down_sample_3d.shape[i]) for i in range(len(down_sample_3d.shape))],
                    "constant", constant_values=-1)

        return vox_map, vox_weights, down_sample_3d

    def pre_process_voxels(self):
        """
        Updates opt_voxels_dict based upon ct to dose_1d voxel map

        :param ct: object if class CT
        :param structs: object of class CT
        :return:
        """
        if self._down_sample_xyz is not None:
            vox_map, vox_weights, down_sample_3d = self.down_sample_voxels(down_sample_xyz=self._down_sample_xyz)
            self.opt_voxels_dict['ct_to_dose_voxel_map'][0] = down_sample_3d
            self._vox_map = vox_map
            self._vox_weights = vox_weights
            for structure_name in self._structs.structures_dict['name']:
                ind = self._structs.structures_dict['name'].index(structure_name)
                # deep copying mask so that it doesnt modify in structures
                vox_3d = deepcopy(self._structs.structures_dict['structure_mask_3d'][ind]) * self.opt_voxels_dict['ct_to_dose_voxel_map'][0]
                # my_plan.structures_dict['voxel_idx'][i] = np.unique(vox_3d[vox_3d > 0])
                vox, counts = np.unique(vox_3d[vox_3d > 0], return_counts=True)
                self.opt_voxels_dict['voxel_idx'][ind] = vox
                self.opt_voxels_dict['voxel_volume_cc'][ind] = counts * np.prod(
                    self._ct.get_ct_res_xyz_mm())/1000  # calculate weight for each voxel
                # self.opt_voxels_dict['voxel_volume_cc'][ind] = counts / np.max(counts)  # calculate weight for each voxel

    @staticmethod
    def sort_beamlets(b_map):
        c = b_map[b_map >= 0]
        c = np.unique(c)
        ind = np.arange(0, len(c))
        c_sort = np.sort(c)
        matrix_ind = [np.where(b_map == c_i) for c_i in c_sort]
        map_copy = b_map.copy()
        for i in range(len(ind)):
            map_copy[matrix_ind[i]] = ind[i]
        return map_copy

    # for voxels idx methods
    def set_opt_voxel_idx(self, plan_obj, structure_name: str) -> None:
        """
        Set opt_voxel_idx for the struct_name based upon ct_dose_voxel_map and struct_name mask

        :param plan_obj: object of class Plan
        :param structure_name: struct_name name
        :return: set the voxel idx in opt_voxels_dict for the struct_name
        """
        ind = plan_obj.structures.structures_dict['name'].index(structure_name)
        vox_3d = plan_obj.structures.structures_dict['structure_mask_3d'][ind] * \
                 self.opt_voxels_dict['ct_to_dose_voxel_map'][0]
        # my_plan.structures_dict['voxel_idx'][i] = np.unique(vox_3d[vox_3d > 0])
        vox, counts = np.unique(vox_3d[vox_3d > 0], return_counts=True)
        self.opt_voxels_dict['voxel_idx'].append(vox)
        # self.opt_voxels_dict['voxel_volume_cc'].append(counts / np.max(counts))  # calculate weight for each voxel
        self.opt_voxels_dict['voxel_volume_cc'].append(
            counts * np.prod(plan_obj.get_ct_res_xyz_mm())/1000)  # calculate weight for each voxel
        self.opt_voxels_dict['name'].append(structure_name)

    def get_fraction_of_vol_in_calc_box(self, structure_name: str):
        return self._structs.get_fraction_of_vol_in_calc_box(structure_name=structure_name)

    def get_opt_voxels_idx(self, structure_name: str) -> np.ndarray:
        """
        Get voxel index for struct_name
        :param structure_name: name of the struct_name in plan
        :return: voxel indexes for the struct_name
        """
        ind = self.opt_voxels_dict['name'].index(structure_name)
        vox_ind = self.opt_voxels_dict['voxel_idx'][ind]
        # vox_ind = np.where(my_plan.opt_voxels_dict['voxel_structure_map'][0][:, ind] == 1)[0]
        return vox_ind

    def get_opt_voxels_volume_cc(self, structure_name: str):
        """
         :param structure_name: name of the struct_name in plan
         :return: voxel size for the struct_name
         """
        ind = self.opt_voxels_dict['name'].index(structure_name)
        vox_weights = self.opt_voxels_dict['voxel_volume_cc'][ind]
        # vox_ind = np.where(my_plan.opt_voxels_dict['voxel_structure_map'][0][:, ind] == 1)[0]
        return vox_weights

    def get_all_beam_ids(self) -> List[int]:
        ids = [self.beamlets_dict[i]['beam_id'] for i in range(len(self.beamlets_dict))]
        return ids



import numpy as np
import sys
import os
import ntpath
import time
import SimpleITK as sitk
from collections import OrderedDict
from . import util
from . import html
if sys.version_info[0] == 2:
    VisdomExceptionBase = Exception
else:
    VisdomExceptionBase = ConnectionError

def convert_visuals_multitask(visuals):
  # Extract the multitask viusals separately and make new dictionary
  visuals_new = OrderedDict()
  for label, image in visuals.items():
    if label == 'real_A':
      visuals_new[label] = image
    else:
      for i in range(image.shape[1]):
        visuals_new[label+str(i)] = image[:,i,:,:,:].unsqueeze(axis=0)
  return visuals_new

def save_images(webpage, visuals, image_path, aspect_ratio=1.0, width=256, npy_out_fnames=['CBCT2CT']):
    """Save images to the disk.

    Parameters:
        webpage (the HTML class) -- the HTML webpage class that stores these imaegs (see html.py for more details)
        visuals (OrderedDict)    -- an ordered dictionary that stores (name, images (either tensor or numpy) ) pairs
        image_path (str)         -- the string is used to create image paths
        aspect_ratio (float)     -- the aspect ratio of saved images
        width (int)              -- the images will be resized to width x width

    This function will save images stored in 'visuals' to the HTML file specified by 'webpage'.
    """
    npz_image_dir = webpage.get_npz_image_dir()
    short_path = ntpath.basename(image_path[0])
    name = os.path.splitext(short_path)[0]
    
    #visuals = convert_visuals_multitask(visuals) # Extract multi task images separately for e.g. CT, Eso label from the channels axis
    
    idx = 0
    for label, im_data in visuals.items():
        
        if 'fake_Dose' in label: # CBCT to CT translated image
            image_numpy = (im_data[0].cpu().float().numpy())[0]  # convert it into a numpy array
            suffix = npy_out_fnames[idx] + '.nrrd'

            image_sitk = sitk.GetImageFromArray(image_numpy)

            save_path = os.path.join(npz_image_dir, name + '_' + suffix)
            sitk.WriteImage(image_sitk, save_path)
            idx += 1
    
class Visualizer():
    def __init__(self, opt):
        # self.opt = opt
        self.display_id = opt.display_id
        self.use_html = opt.isTrain and not opt.no_html
        self.win_size = opt.display_winsize
        self.name = opt.name
        self.saved = False
        if self.display_id > 0:
            import visdom
            self.vis = visdom.Visdom(port = opt.display_port)
            self.display_single_pane_ncols = opt.display_ncols

        if self.use_html:
            self.web_dir = os.path.join(opt.checkpoints_dir, opt.name, 'web')
            self.img_dir = os.path.join(self.web_dir, 'images')
            print('create web directory %s...' % self.web_dir)
            util.mkdirs([self.web_dir, self.img_dir])
        self.log_name = os.path.join(opt.checkpoints_dir, opt.name, 'loss_log.txt')
        with open(self.log_name, "a") as log_file:
            now = time.strftime("%c")
            log_file.write('================ Training Loss (%s) ================\n' % now)

    # |visuals|: dictionary of images to display or save
    def display_current_results(self, visuals, epoch, save_results=False):
        #if self.display_id > 0: # show images in the browser
        #visuals = convert_visuals_multitask(visuals)
        if False:
            if self.display_single_pane_ncols > 0:
                h, w = next(iter(visuals.values())).shape[:2]
                table_css = """<style>
    table {border-collapse: separate; border-spacing:4px; white-space:nowrap; text-align:center}
    table td {width: %dpx; height: %dpx; padding: 4px; outline: 4px solid black}
</style>""" % (w, h)
                ncols = self.display_single_pane_ncols
                title = self.name
                label_html = ''
                label_html_row = ''
                nrows = int(np.ceil(len(visuals.items()) / ncols))
                images = []
                idx = 0
                for label, image_numpy in visuals.items():
                    label_html_row += '<td>%s</td>' % label
                    images.append(image_numpy.transpose([2, 0, 1]))
                    idx += 1
                    if idx % ncols == 0:
                        label_html += '<tr>%s</tr>' % label_html_row
                        label_html_row = ''
                white_image = np.ones_like(image_numpy.transpose([2, 0, 1]))*255
                while idx % ncols != 0:
                    images.append(white_image)
                    label_html_row += '<td></td>'
                    idx += 1
                if label_html_row != '':
                    label_html += '<tr>%s</tr>' % label_html_row
                # pane col = image row
                self.vis.images(images, nrow=ncols, win=self.display_id + 1,
                                padding=2, opts=dict(title=title + ' images'))
                label_html = '<table>%s</table>' % label_html
                self.vis.text(table_css + label_html, win = self.display_id + 2,
                              opts=dict(title=title + ' labels'))
            else:
                idx = 1
                for label, image_numpy in visuals.items():
                    #image_numpy = np.flipud(image_numpy)
                    self.vis.image(image_numpy.transpose([2,0,1]), opts=dict(title=label),
                                       win=self.display_id + idx)
                    idx += 1

        if self.use_html: # save images to a html file
            for label, image_numpy in visuals.items():
                image_numpy = util.tensor2im3d(image_numpy.data)
                #print(image_numpy.shape)
                if len(image_numpy.shape) == 4:
                    if 'fake_Dose' in label or 'real_Dose' in label:

                        img_path = os.path.join(self.img_dir, 'epoch%.3d_%s.nrrd' % (epoch, label))
                        image_sitk = sitk.GetImageFromArray(image_numpy[0])
                        sitk.WriteImage(image_sitk, img_path)
                else:
                    img_path = os.path.join(self.img_dir, 'epoch%.3d_%s.png' % (epoch, label))
                    util.save_image(image_numpy, img_path)
            # update website
            webpage = html.HTML(self.web_dir, 'Experiment name = %s' % self.name, refresh=1)
            for n in range(epoch, 0, -1):
                webpage.add_header('epoch [%d]' % n)
                ims = []
                txts = []
                links = []

                for label, image_numpy in visuals.items():
                    img_path = 'epoch%.3d_%s.gif' % (n, label)
                    ims.append(img_path)
                    txts.append(label)
                    links.append(img_path)
                webpage.add_images(ims, txts, links, width=self.win_size)
            webpage.save()
    def reset(self):
        """Reset the self.saved status"""
        self.saved = False
    # errors: dictionary of error labels and values
    def plot_current_losses(self, epoch, counter_ratio, losses):
        """display the current losses on visdom display: dictionary of error labels and values

        Parameters:
            epoch (int)           -- current epoch
            counter_ratio (float) -- progress (percentage) in the current epoch, between 0 to 1
            losses (OrderedDict)  -- training losses stored in the format of (name, float) pairs
        """
        if not hasattr(self, 'plot_data'):
            self.plot_data = {'X': [], 'Y': [], 'legend': list(losses.keys())}
        self.plot_data['X'].append(epoch + counter_ratio)
        self.plot_data['Y'].append([losses[k] for k in self.plot_data['legend']])
        try:
            self.vis.line(
                X=np.stack([np.array(self.plot_data['X'])] * len(self.plot_data['legend']), 1),
                Y=np.array(self.plot_data['Y']),
                opts={
                    'title': self.name + ' loss over time',
                    'legend': self.plot_data['legend'],
                    'xlabel': 'epoch',
                    'ylabel': 'loss'},
                win=self.display_id)
        except VisdomExceptionBase:
            self.create_visdom_connections()

    # losses: same format as |losses| of plot_current_losses
    def print_current_losses(self, epoch, iters, losses, t_comp, t_data):
        """print current losses on console; also save the losses to the disk

        Parameters:
            epoch (int) -- current epoch
            iters (int) -- current training iteration during this epoch (reset to 0 at the end of every epoch)
            losses (OrderedDict) -- training losses stored in the format of (name, float) pairs
            t_comp (float) -- computational time per data point (normalized by batch_size)
            t_data (float) -- data loading time per data point (normalized by batch_size)
        """
        message = '(epoch: %d, iters: %d, time: %.3f, data: %.3f) ' % (epoch, iters, t_comp, t_data)
        for k, v in losses.items():
            message += '%s: %.3f ' % (k, v)

        print(message)  # print the message
        with open(self.log_name, "a") as log_file:
            log_file.write('%s\n' % message)  # save the message

    # save image to the disk
    def save_images(self, webpage, visuals, image_path):
        image_dir = webpage.get_image_dir()
        short_path = ntpath.basename(image_path[0])
        name = os.path.splitext(short_path)[0]

        webpage.add_header(name)
        ims = []
        txts = []
        links = []

        for label, image_numpy in visuals.items():
            image_name = '%s_%s.png' % (name, label)
            save_path = os.path.join(image_dir, image_name)
            util.save_image(image_numpy, save_path)

            ims.append(image_name)
            txts.append(label)
            links.append(image_name)
        webpage.add_images(ims, txts, links, width=self.win_size)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug 11 13:16:11 2020

@author: ndahiya
"""

import os.path
from data.base_dataset import BaseDataset, get_params, get_transform, transform_3d_data
from data.image_folder import make_dataset
import numpy as np
import torch


class DosePred3DDataset(BaseDataset):
    """
  A dataset class for 2D CT to Dose prediction dataset.
  
  It assumes that the directory '/path/to/data/train' contains *.npz images which
  have at least two arrays named 'CT', 'DOSE'. Assuming data is already the desired size.
  
  During test time, you need to prepare a directory '/path/to/data/test'.
  """

    def __init__(self, opt):
        """Initialize this dataset class.

    Parameters:
    opt (Option class) -- stores all the experiment flags; needs to be a subclass of BaseOptions
    """
        BaseDataset.__init__(self, opt)
        self.dir_AB = os.path.join(opt.dataroot, opt.phase)  # get the image directory
        self.AB_paths = sorted(make_dataset(self.dir_AB, opt.max_dataset_size))  # get image paths
        assert (self.opt.load_size >= self.opt.crop_size)  # crop_size should be smaller than the size of loaded image
        self.input_nc = self.opt.output_nc if self.opt.direction == 'BtoA' else self.opt.input_nc
        self.output_nc = self.opt.input_nc if self.opt.direction == 'BtoA' else self.opt.output_nc
        self.phase = self.opt.phase

    def __getitem__(self, index):
        """Return a data point and its metadata information.

    Parameters:
      index - - a random integer for data indexing

    Returns a dictionary that contains A, B, A_paths and B_paths
      A (tensor) - - an image in the input domain
      B (tensor) - - its corresponding image in the target domain
      A_paths (str) - - image paths
      B_paths (str) - - image paths (same as A_paths)
      """
        # read a image given a random integer index
        AB_path = self.AB_paths[index]
        AB = np.load(AB_path)  ##Added (Gourav)

        # Get AB image into A and B
        A = AB['CT']
        B = AB['DOSE']
        OAR = AB['OAR']
        PTV = AB['PTV']
        # hist = AB['HIST']
        # bins = AB['BINS']
        beam = AB['BEAM']

        # apply the same transform to both A and B
        if self.phase == 'train':
            # A, B, OAR, beam, hist, bins = transform_3d_data(A, B, OAR, PTV, beam, hist, bins, transform=False)
            A, B, OAR, beam = transform_3d_data(A, B, OAR, PTV, beam, transform=False)
            A = torch.unsqueeze(A, dim=0)  # Add channel dimensions as data is 3D
            B = torch.unsqueeze(B, dim=0)
            beam = torch.unsqueeze(beam, dim=0)

            A = torch.cat((A, beam, OAR), axis=0)
            # A = torch.cat((A, OAR), axis=0)  # No Beam
            # print(OAR.shape, hist.shape, bins.shape)
        else:
            # A, B, OAR, beam, hist, bins = transform_3d_data(A, B, OAR, PTV, beam, hist, bins, transform=False)
            A, B, OAR, beam = transform_3d_data(A, B, OAR, PTV, beam, transform=False)
            A = torch.unsqueeze(A, dim=0)  # Add channel dimensions as data is 3D
            B = torch.unsqueeze(B, dim=0)
            beam = torch.unsqueeze(beam, dim=0)
            A = torch.cat((A, beam, OAR), axis=0)
            # A = torch.cat((A, OAR), axis=0)  # No beam
            # print(OAR.max())
        return {'A': A, 'B': B, 'A_paths': AB_path, 'B_paths': AB_path} #'HIST': hist, 'BINS': bins}

    def __len__(self):
        """Return the total number of images in the dataset."""
        return len(self.AB_paths)



"""General-purpose training script for image-to-image translation.

This script works for various models (with option '--model': e.g., pix2pix, cyclegan, colorization) and
different datasets (with option '--dataset_mode': e.g., aligned, unaligned, single, colorization).
You need to specify the dataset ('--dataroot'), experiment name ('--name'), and model ('--model').

It first creates model, dataset, and visualizer given the option.
It then does standard network training. During the training, it also visualize/save the images, print/save the loss plot, and save models.
The script supports continue/resume training. Use '--continue_train' to resume your previous training.

Example:
    Train a CycleGAN model:
        python train.py --dataroot ./datasets/maps --name maps_cyclegan --model cycle_gan
    Train a pix2pix model:
        python train.py --dataroot ./datasets/facades --name facades_pix2pix --model pix2pix --direction BtoA

See options/base_options.py and options/train_options.py for more training options.
See training and test tips at: https://github.com/junyanz/pytorch-CycleGAN-and-pix2pix/blob/master/docs/tips.md
See frequently asked questions at: https://github.com/junyanz/pytorch-CycleGAN-and-pix2pix/blob/master/docs/qa.md
"""
import time
from options.train_options import TrainOptions
from data import create_dataset
from models import create_model
from util.visualizer import Visualizer
import matplotlib.pyplot as plt
import os
import pandas as pd
import torch
import numpy as np

if __name__ == '__main__':

    torch.cuda.empty_cache()
    opt = TrainOptions().parse()  # get training options
    dataset = create_dataset(opt)  # create a dataset given opt.dataset_mode and other options
    dataset_size = len(dataset)  # get the number of images in the dataset.
    print('The number of training images = %d' % dataset_size)
    print('shuffle :{}'.format(not opt.serial_batches))
    model = create_model(opt)  # create a model given opt.model and other options
    # model.print_networks(verbose=True)
    model.setup(opt)  # regular setup: load and print networks; create schedulers
    visualizer = Visualizer(opt)  # create a visualizer that display/save images and plots
    total_iters = 0  # the total number of training iterations
    total_time_start = time.time()
    loss_summary = []
    loss_summary_test = []

    # ##Create test dataset for evaluation of test loss
    optTest = opt
    optTest.phase = 'test'
    optTest.num_threads = 0  # test code only supports num_threads = 1
    optTest.batch_size = 1  # test code only supports batch_size = 1
    optTest.serial_batches = True  # disable data shuffling; comment this line if results on randomly chosen images are needed.
    optTest.no_flip = True  # no flip; comment this line if results on flipped images are needed.
    optTest.display_id = -1  # no visdom display; the test code saves the results to a HTML file.
    optTest.isTrain = False
    optTest.mode = 'eval'
    dataset_test = create_dataset(optTest)
    dataset_size_test = len(dataset_test)  # get the number of images in the dataset.
    print('The number of validation images = %d' % dataset_size_test)

    for epoch in range(opt.epoch_count,
                       opt.n_epochs + opt.n_epochs_decay + 1):  # outer loop for different epochs; we save the model by <epoch_count>, <epoch_count>+<save_latest_freq>
        epoch_start_time = time.time()  # timer for entire epoch
        iter_data_time = time.time()  # timer for data loading per iteration
        epoch_iter = 0  # the number of training iterations in current epoch, reset to 0 every epoch
        visualizer.reset()  # reset the visualizer: make sure it saves the results to HTML at least once every epoch

        model.set_epoch(epoch)

        total_loss_each_epoch = 0
        for i, data in enumerate(dataset):  # inner loop within one epoch
            iter_start_time = time.time()  # timer for computation per iteration
            if total_iters % opt.print_freq == 0:
                t_data = iter_start_time - iter_data_time

            total_iters += opt.batch_size
            epoch_iter += opt.batch_size
            model.set_input(data)  # unpack data from dataset and apply preprocessing
            # if i == 0:
            #     model.save_weights(opt)
            model.optimize_parameters()  # calculate loss functions, get gradients, update network weights

            if total_iters % opt.display_freq == 0:  # display images on visdom and save images to a HTML file
                save_result = total_iters % opt.update_html_freq == 0
                model.compute_visuals()
                visualizer.display_current_results(model.get_current_visuals(), epoch, save_result)

            if total_iters % opt.print_freq == 0:  # print training losses and save logging information to the disk
                losses = model.get_current_losses()
                total_loss_each_epoch = total_loss_each_epoch + sum(losses.values())
                t_comp = (time.time() - iter_start_time) / opt.batch_size
                visualizer.print_current_losses(epoch, epoch_iter, losses, t_comp, t_data)
                # if opt.display_id > 0:
                #     visualizer.plot_current_losses(epoch, float(epoch_iter) / dataset_size, losses)

            if total_iters % opt.save_latest_freq == 0:  # cache our latest model every <save_latest_freq> iterations
                print('saving the latest model (epoch %d, total_iters %d)' % (epoch, total_iters))
                save_suffix = 'iter_%d' % total_iters if opt.save_by_iter else 'latest'
                model.save_networks(save_suffix)

            iter_data_time = time.time()
        if epoch % opt.save_epoch_freq == 0:  # cache our model every <save_epoch_freq> epochs
            print('saving the model at the end of epoch %d, iters %d' % (epoch, total_iters))
            model.save_networks('latest')
            model.save_networks(epoch)

        print('End of epoch %d / %d \t Time Taken: %d sec' % (epoch, opt.n_epochs + opt.n_epochs_decay, time.time() - epoch_start_time))
        model.update_learning_rate()  # update learning rates at the end of every epoch.
        avg_loss_each_epoch_train = total_loss_each_epoch/len(dataset)
        print('Avg Loss at the end of epoch {}:{}'.format(epoch, avg_loss_each_epoch_train))
        loss_summary.append(avg_loss_each_epoch_train)

        #Evaluating Validation Loss
        total_loss_each_epoch_test = 0
        model.eval()
        for i, data_test in enumerate(dataset_test):
            # if i >= opt.num_test:  # only apply our model to opt.num_test images.
            #   break
            with torch.no_grad():
                model.set_input(data_test)  # unpack data from data loader
                model.test()
                model.calculate_validation_loss()
                val_losses = model.get_current_losses()
                # print('Validation loss for iter {} is {}'.format(i+1, val_losses))
            total_loss_each_epoch_test = total_loss_each_epoch_test + sum(val_losses.values())
        avg_loss_each_epoch_test = total_loss_each_epoch_test / len(dataset_test)
        print('Avg Test Loss at the end of epoch {}:{}'.format(epoch, avg_loss_each_epoch_test))
        loss_summary_test.append(avg_loss_each_epoch_test)
        ##Reset to training mode
        model.train()



    # plt.plot(losssummary, np.arange(opt.epoch_count, opt.n_epochs + opt.n_epochs_decay, 1), 'r--')
    total_time_end = time.time()
    total_time = total_time_end - total_time_start
    print('Total Time for training is {}'.format(total_time))

    ##print total loss after each epoch
    loss_dir = os.path.join(opt.checkpoints_dir, opt.name)
    loss_filename = 'loss_log.txt'
    loss_filename = os.path.join(loss_dir, loss_filename)
    with open(loss_filename, 'a') as f:
        print('Total Time for training is {}'.format(total_time), file=f)

    loss_filename = 'total_loss_train.txt'
    loss_filename = os.path.join(loss_dir, loss_filename)
    # print(loss_summary)
    print(loss_filename)
    with open(loss_filename, 'w') as f:
        for i, line in enumerate(loss_summary):
            print("{} {}".format(i + 1, line), file=f)
    df = pd.read_csv(loss_filename, delim_whitespace=True, header=None)
    df.columns = ["Epoch", "Train_Loss"]
    df.to_excel(r"{}.xlsx".format(loss_filename))

    loss_filename = 'total_loss_test.txt'
    loss_filename = os.path.join(loss_dir, loss_filename)
    # print(loss_summary)
    print(loss_filename)
    with open(loss_filename, 'w') as f:
        for i, line in enumerate(loss_summary_test):
            print("{} {}".format(i + 1, line), file=f)
    df = pd.read_csv(loss_filename, delim_whitespace=True, header=None)
    df.columns = ["Epoch", "Test_Loss"]
    df.to_excel(r"{}.xlsx".format(loss_filename))



import numpy as np


def view_in_slicer_jupyter(my_plan, dose_1d: np.ndarray = None, sol: dict = None,
                           ct_name: str = 'ct', dose_name: str = 'dose', struct_set_name: str = 'rt_struct',
                           show_ct: bool = True,
                           show_dose: bool = True,
                           show_structs: bool = True):
    """
    This method helps to visualize CT, Dose and Rt struct in 3d slicer


    :param my_plan: object of class plan
    :param dose_1d: dose in 1d
    :param sol: solution dictionary
    :param ct_name: Default to 'ct'. name of the ct node in 3D slicer
    :param dose_name: Default to 'dose'. name of the dose node in 3D slicer
    :param struct_set_name: name of the rtstruct
    :param show_structs: default to True. If false, will not create struct_name node
    :param show_dose: default to True. If false, will not create dose node
    :param show_ct:default to True. If false, will not create ct node
    :return: visualize in slicer jupyter
    """
    try:
        import slicer
    except ImportError:
        print('Please use slicer kernel before using this method')

    if show_ct:
        slicer.mrmlScene.AddNewNodeByClass("vtkMRMLScalarVolumeNode", ct_name)
        ct_node = slicer.util.getNode(ct_name)

        ct_node.SetOrigin(list(np.array(my_plan.ct.ct_dict['origin_xyz_mm']) * np.array(
            [-1, -1, 1])))  # convert to RAS origin. -ve in x and y direction
        ct_node.SetSpacing(my_plan.ct.ct_dict['resolution_xyz_mm'])
        ct_node.SetIJKToRASDirections(np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]]))  # set itk to RAS direction
        slicer.util.updateVolumeFromArray(ct_node, my_plan.ct.ct_dict['ct_hu_3d'][0])
        slicer.util.setSliceViewerLayers(background=ct_node)

    # Another way to push simple itk image
    # import sitkUtils
    # ct = sitk.GetImageFromArray(my_plan.ct['ct_hu_3d'][0])
    # ct.SetOrigin(my_plan.ct['origin_xyz_mm'])
    # ct.SetSpacing(my_plan.ct['resolution_xyz_mm'])
    # sitkUtils.PushVolumeToSlicer(ct, targetNode='ct')

    # plot dose
    if show_dose:
        if dose_1d is not None:
            dose_arr = my_plan.inf_matrix.dose_1d_to_3d(dose_1d=dose_1d)
        else:
            dose_1d = sol['inf_matrix'].A * sol['optimal_intensity'] * my_plan.get_num_of_fractions()
            dose_arr = sol['inf_matrix'].dose_1d_to_3d(dose_1d=dose_1d)
        slicer.mrmlScene.AddNewNodeByClass("vtkMRMLScalarVolumeNode", dose_name)
        dose_node = slicer.util.getNode(dose_name)
        dose_node.SetOrigin(list(np.array(my_plan.ct.ct_dict['origin_xyz_mm']) * np.array([-1, -1, 1])))
        dose_node.SetSpacing(my_plan.ct.ct_dict['resolution_xyz_mm'])
        dose_node.SetIJKToRASDirections(np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]]))
        slicer.util.updateVolumeFromArray(dose_node, dose_arr)
        slicer.util.setSliceViewerLayers(foreground=dose_node)
        slicer.util.setSliceViewerLayers(foregroundOpacity=0.4)
        # dose_node.GetVolumeDisplayNode.

    # plot structures
    if show_structs:
        structure_node = slicer.mrmlScene.AddNewNodeByClass("vtkMRMLSegmentationNode", struct_set_name)
        for i, struct_name in enumerate(my_plan.structures.structures_dict['name']):
            print('Importing struct_name : ' + struct_name)
            slicer.util.addVolumeFromArray(my_plan.structures.structures_dict['structure_mask_3d'][i] * (i + 1),
                                           name=struct_name, nodeClassName="vtkMRMLLabelMapVolumeNode")
            lmap = slicer.util.getNode(struct_name)
            lmap.SetOrigin(list(np.array(my_plan.ct.ct_dict['origin_xyz_mm']) * np.array([-1, -1, 1])))
            lmap.SetSpacing(my_plan.ct.ct_dict['resolution_xyz_mm'])
            lmap.SetIJKToRASDirections(np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]]))
            slicer.modules.segmentations.logic().ImportLabelmapToSegmentationNode(lmap, structure_node)
            slicer.mrmlScene.RemoveNode(lmap)
            segId = structure_node.GetSegmentation().GetNthSegment(i)
            segId.SetName(struct_name)



import sys
import os

print(sys.argv[1])
print(sys.argv[2])
img_dir = sys.argv[1]
# print(patient_folder_path)
# ct = sitk.ReadImage(os.path.join(patient_folder_path, "ct.nrrd"))
# ct_arr = sitk.GetArrayFromImage(ct)
# slicer.mrmlScene.AddNewNodeByClass("vtkMRMLScalarVolumeNode", 'ct')
slicer.util.loadVolume(os.path.join(img_dir, "ct.nrrd"))
ct_node = slicer.util.getNode('ct')
slicer.util.setSliceViewerLayers(background=ct_node)

# slicer.mrmlScene.AddNewNodeByClass("vtkMRMLScalarVolumeNode", 'dose_1d')
slicer.util.loadVolume(os.path.join(img_dir, "dose.nrrd"))
doseNode = slicer.util.getNode('dose')
slicer.util.setSliceViewerLayers(foreground=doseNode)
slicer.util.setSliceViewerLayers(foregroundOpacity=0.4)

# structure_node = slicer.mrmlScene.AddNewNodeByClass("vtkMRMLSegmentationNode", 'rt_struct')
slicer.util.loadSegmentation(os.path.join(img_dir, "rtss.seg.nrrd"))
rtss = slicer.util.getNode('rtss')
segmentation = rtss.GetSegmentation()
# n = segmentation.GetNumberOfSegments()
struct_list = sys.argv[2].split(',')
# print(struct_list)
# struct_list = ['PTV', 'CTV', 'BLAD_WALL', 'BLADDER_TRIGONE', 'FEMUR_L', 'FEMUR_R', 'RECT_WALL', 'URETHRA', 'SKIN', 'RIND_0', 'RIND_1', 'RIND_2', 'RIND_3', 'RIND_4']
for i in range(len(struct_list)):
    segment = segmentation.GetNthSegment(i)
    segment.SetName(struct_list[i])

# struct_name node



import torch
import torch.nn as nn
from torch.nn import init
import functools
from torch.optim import lr_scheduler
import numpy as np


###############################################################################
# Functions
###############################################################################

class Identity(nn.Module):
    def forward(self, x):
        return x


def init_weights(net, init_type='normal', init_gain=0.02):
    """Initialize network weights.

    Parameters:
        net (network)   -- network to be initialized
        init_type (str) -- the name of an initialization method: normal | xavier | kaiming | orthogonal
        init_gain (float)    -- scaling factor for normal, xavier and orthogonal.

    We use 'normal' in the original pix2pix and CycleGAN paper. But xavier and kaiming might
    work better for some applications. Feel free to try yourself.
    """

    def init_func(m):  # define the initialization function
        classname = m.__class__.__name__
        if hasattr(m, 'weight') and (classname.find('Conv') != -1 or classname.find('Linear') != -1):
            if init_type == 'normal':
                init.normal_(m.weight.data, 0.0, init_gain)
            elif init_type == 'xavier':
                init.xavier_normal_(m.weight.data, gain=init_gain)
            elif init_type == 'kaiming':
                init.kaiming_normal_(m.weight.data, a=0, mode='fan_in')
            elif init_type == 'orthogonal':
                init.orthogonal_(m.weight.data, gain=init_gain)
            else:
                raise NotImplementedError('initialization method [%s] is not implemented' % init_type)
            if hasattr(m, 'bias') and m.bias is not None:
                init.constant_(m.bias.data, 0.0)
        elif classname.find(
                'BatchNorm3d') != -1:  # BatchNorm Layer's weight is not a matrix; only normal distribution applies.
            init.normal_(m.weight.data, 1.0, init_gain)
            init.constant_(m.bias.data, 0.0)

    print('initialize network with %s' % init_type)
    net.apply(init_func)  # apply the initialization function <init_func>


def get_norm_layer(norm_type='instance'):
    """Return a normalization layer

    Parameters:
        norm_type (str) -- the name of the normalization layer: batch | instance | none

    For BatchNorm, we use learnable affine parameters and track running statistics (mean/stddev).
    For InstanceNorm, we do not use learnable affine parameters. We do not track running statistics.
    """
    if norm_type == 'batch':
        norm_layer = functools.partial(nn.BatchNorm3d, affine=True, track_running_stats=True)
    elif norm_type == 'instance':
        norm_layer = functools.partial(nn.InstanceNorm3d, affine=False, track_running_stats=False)
    elif norm_type == 'none':
        def norm_layer(x):
            return Identity()
    else:
        raise NotImplementedError('normalization layer [%s] is not found' % norm_type)
    return norm_layer


def get_scheduler(optimizer, opt):
    """Return a learning rate scheduler

    Parameters:
        optimizer          -- the optimizer of the network
        opt (option class) -- stores all the experiment flags; needs to be a subclass of BaseOptions．　
                              opt.lr_policy is the name of learning rate policy: linear | step | plateau | cosine

    For 'linear', we keep the same learning rate for the first <opt.n_epochs> epochs
    and linearly decay the rate to zero over the next <opt.n_epochs_decay> epochs.
    For other schedulers (step, plateau, and cosine), we use the default PyTorch schedulers.
    See https://pytorch.org/docs/stable/optim.html for more details.
    """
    if opt.lr_policy == 'linear':
        def lambda_rule(epoch):
            lr_l = 1.0 - max(0, epoch + opt.epoch_count - opt.n_epochs) / float(opt.n_epochs_decay + 1)
            return lr_l

        scheduler = lr_scheduler.LambdaLR(optimizer, lr_lambda=lambda_rule)
    elif opt.lr_policy == 'step':
        scheduler = lr_scheduler.StepLR(optimizer, step_size=opt.lr_decay_iters, gamma=0.1)
    elif opt.lr_policy == 'plateau':
        scheduler = lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', factor=0.2, threshold=0.01, patience=5)
    elif opt.lr_policy == 'cosine':
        scheduler = lr_scheduler.CosineAnnealingLR(optimizer, T_max=opt.n_epochs, eta_min=0)
    else:
        return NotImplementedError('learning rate policy [%s] is not implemented', opt.lr_policy)
    return scheduler


def init_net(net, init_type='normal', init_gain=0.02, gpu_ids=[]):
    """Initialize a network: 1. register CPU/GPU device (with multi-GPU support); 2. initialize the network weights
    Parameters:
        net (network)      -- the network to be initialized
        init_type (str)    -- the name of an initialization method: normal | xavier | kaiming | orthogonal
        gain (float)       -- scaling factor for normal, xavier and orthogonal.
        gpu_ids (int list) -- which GPUs the network runs on: e.g., 0,1,2

    Return an initialized network.
    """
    if len(gpu_ids) > 0:
        assert (torch.cuda.is_available())
        net.to(gpu_ids[0])
        net = torch.nn.DataParallel(net, gpu_ids)  # multi-GPUs
    init_weights(net, init_type, init_gain=init_gain)
    return net


def define_G(input_nc, output_nc, ngf, netG, norm='batch', use_dropout=False, init_type='normal', init_gain=0.02,
             gpu_ids=[]):
    net = None
    norm_layer = get_norm_layer(norm_type=norm)

    if netG == 'unet_128':
        net = UnetGenerator3d(input_nc, output_nc, 7, ngf, norm_layer=norm_layer, use_dropout=use_dropout)
    elif netG == 'unet_256':
        net = UnetGenerator3d(input_nc, output_nc, 8, ngf, norm_layer=norm_layer, use_dropout=use_dropout)
    elif netG == 'stand_unet':
        net = UNet3D(in_ch=input_nc, out_ch=output_nc)
    else:
        raise NotImplementedError('Generator model name [%s] is not recognized' % netG)

    return init_net(net, init_type, init_gain, gpu_ids)


def print_network(net):
    num_params = 0
    for param in net.parameters():
        num_params += param.numel()
    print(net)
    print('Total number of parameters: %d' % num_params)


##############################################################################
# Classes
##############################################################################

# 3D version of UnetGenerator
class UnetGenerator3d(nn.Module):
    def __init__(self, input_nc, output_nc, num_downs, ngf=64, norm_layer=nn.BatchNorm3d, use_dropout=False):
        """Construct a 3D Unet generator
        Parameters:
            input_nc (int)  -- the number of channels in input images
            output_nc (int) -- the number of channels in output images
            num_downs (int) -- the number of downsamplings in UNet. For example, # if |num_downs| == 7,
                                image of size 128x128 will become of size 1x1 # at the bottleneck
            ngf (int)       -- the number of filters in the last conv layer
            norm_layer      -- normalization layer

        We construct the U-Net from the innermost layer to the outermost layer.
        It is a recursive process.
        """
        super(UnetGenerator3d, self).__init__()
        # construct unet structure
        unet_block = UnetSkipConnectionBlock3d(ngf * 8, ngf * 8, input_nc=None, submodule=None, norm_layer=norm_layer,
                                               innermost=True)  # add the innermost layer
        for i in range(num_downs - 5):  # add intermediate layers with ngf * 8 filters
            unet_block = UnetSkipConnectionBlock3d(ngf * 8, ngf * 8, input_nc=None, submodule=unet_block,
                                                   norm_layer=norm_layer, use_dropout=use_dropout)
        # gradually reduce the number of filters from ngf * 8 to ngf
        unet_block = UnetSkipConnectionBlock3d(ngf * 4, ngf * 8, input_nc=None, submodule=unet_block,
                                               norm_layer=norm_layer)
        unet_block = UnetSkipConnectionBlock3d(ngf * 2, ngf * 4, input_nc=None, submodule=unet_block,
                                               norm_layer=norm_layer)
        unet_block = UnetSkipConnectionBlock3d(ngf, ngf * 2, input_nc=None, submodule=unet_block, norm_layer=norm_layer)
        self.model = UnetSkipConnectionBlock3d(output_nc, ngf, input_nc=input_nc, submodule=unet_block, outermost=True,
                                               norm_layer=norm_layer)  # add the outermost layer

    def forward(self, input):
        """Standard forward"""
        return self.model(input)


class UnetSkipConnectionBlock3d(nn.Module):
    """Defines the Unet submodule with skip connection.
        X -------------------identity----------------------
        |-- downsampling -- |submodule| -- upsampling --|
    """

    def __init__(self, outer_nc, inner_nc, input_nc=None,
                 submodule=None, outermost=False, innermost=False, norm_layer=nn.BatchNorm3d, use_dropout=False):
        """Construct a Unet submodule with skip connections.

        Parameters:
            outer_nc (int) -- the number of filters in the outer conv layer
            inner_nc (int) -- the number of filters in the inner conv layer
            input_nc (int) -- the number of channels in input images/features
            submodule (UnetSkipConnectionBlock) -- previously defined submodules
            outermost (bool)    -- if this module is the outermost module
            innermost (bool)    -- if this module is the innermost module
            norm_layer          -- normalization layer
            use_dropout (bool)  -- if use dropout layers.
        """
        super(UnetSkipConnectionBlock3d, self).__init__()
        self.outermost = outermost
        if type(norm_layer) == functools.partial:
            use_bias = norm_layer.func == nn.InstanceNorm3d
        else:
            use_bias = norm_layer == nn.InstanceNorm3d
        if input_nc is None:
            input_nc = outer_nc
        downconv = nn.Conv3d(input_nc, inner_nc, kernel_size=4,
                             stride=2, padding=1, bias=use_bias)
        downrelu = nn.LeakyReLU(0.2, True)
        downnorm = norm_layer(inner_nc)
        uprelu = nn.ReLU(True)
        upnorm = norm_layer(outer_nc)

        interp = 'trilinear'  # 'nearest'
        transp_conv = False  # Use transposed convolution or resize convolution?
        # transp_conv = True

        if outermost:
            if transp_conv is True:
                upconv = [nn.ConvTranspose3d(inner_nc * 2, outer_nc,
                                             kernel_size=4, stride=2,
                                             padding=1)]
            else:
                upsamp = nn.Upsample(scale_factor=2, mode=interp)
                conv = nn.Conv3d(inner_nc * 2, outer_nc, kernel_size=3, stride=1, padding=1)
                upconv = [upsamp, conv]

            down = [downconv]
            up = [uprelu, *upconv, nn.ReLU()]

            model = down + [submodule] + up
        elif innermost:
            if transp_conv is True:
                upconv = [nn.ConvTranspose3d(inner_nc, outer_nc,
                                             kernel_size=4, stride=2,
                                             padding=1, bias=use_bias)]
            else:
                upsamp = nn.Upsample(scale_factor=2, mode=interp)
                conv = nn.Conv3d(inner_nc, outer_nc, kernel_size=3, stride=1, padding=1, bias=use_bias)
                upconv = [upsamp, conv]

            down = [downrelu, downconv]
            up = [uprelu, *upconv, upnorm]
            model = down + up
        else:
            if transp_conv is True:
                upconv = [nn.ConvTranspose3d(inner_nc * 2, outer_nc,
                                             kernel_size=4, stride=2,
                                             padding=1, bias=use_bias)]
            else:
                upsamp = nn.Upsample(scale_factor=2, mode=interp)
                conv = nn.Conv3d(inner_nc * 2, outer_nc, kernel_size=3, stride=1, padding=1, bias=use_bias)
                upconv = [upsamp, conv]
            down = [downrelu, downconv, downnorm]
            up = [uprelu, *upconv, upnorm]

            if use_dropout:
                model = down + [submodule] + up + [nn.Dropout(0.5)]
            else:
                model = down + [submodule] + up

        self.model = nn.Sequential(*model)

    def forward(self, x):
        if self.outermost:
            return self.model(x)
        else:  # add skip connections
            return torch.cat([x, self.model(x)], 1)


# Standard 3D Unet with padding; No dropout
class Block(nn.Module):
    def __init__(self, in_ch, out_ch):
        super().__init__()
        self.conv1 = nn.Conv3d(in_ch, out_ch, kernel_size=3, padding=1)
        self.relu = nn.ReLU()
        self.conv2 = nn.Conv3d(out_ch, out_ch, kernel_size=3, padding=1)
        self.BN = nn.BatchNorm3d(out_ch)
        # block = [conv1, relu, BN, conv2, relu, BN]
        # self.model = nn.Sequential(*block)

    def forward(self, x):
        return self.BN(self.relu(self.conv2(self.BN(self.relu(self.conv1(x))))))
        # return self.model(x)


class Encoder(nn.Module):
    def __init__(self, in_ch, chs=(32, 64, 128, 256, 512)):
        super().__init__()
        chs = (in_ch,) + chs
        self.enc_blocks = nn.ModuleList([Block(chs[i], chs[i + 1]) for i in range(len(chs) - 1)])
        self.pool = nn.MaxPool3d(2)

    def forward(self, x):
        ftrs = []
        for block in self.enc_blocks:
            x = block(x)
            ftrs.append(x)
            x = self.pool(x)
        return ftrs


class ResizeConv(nn.Module):
    # Resize (upsample) with nearest neighbor or trilinear interpolation and then do convolution
    def __init__(self, in_ch, out_ch, interp='nearest'):
        super().__init__()
        self.upsamp = nn.Upsample(scale_factor=2, mode=interp)
        self.conv = nn.Conv3d(in_ch, out_ch, kernel_size=3, stride=1, padding=1)

    def forward(self, x):
        x = self.upsamp(x)
        x = self.conv(x)
        return x


class Decoder(nn.Module):
    def __init__(self, chs=(1024, 512, 256, 128, 64)):
        super().__init__()
        self.chs = chs
        # self.upconvs = nn.ModuleList([nn.ConvTranspose3d(chs[i], chs[i + 1], 2, 2) for i in range(len(chs) - 1)])
        self.upconvs = nn.ModuleList([ResizeConv(chs[i], chs[i + 1], interp='trilinear') for i in range(len(chs) - 1)])
        self.dec_blocks = nn.ModuleList([Block(chs[i], chs[i + 1]) for i in range(len(chs) - 1)])

    def forward(self, x, encoder_features):
        for i in range(len(self.chs) - 1):
            x = self.upconvs[i](x)
            # enc_ftrs = self.crop(encoder_features[i], x)
            x = torch.cat([x, encoder_features[i]], dim=1)
            x = self.dec_blocks[i](x)
        return x

    # def crop(self, enc_ftrs, x):
    #     _, _, H, W = x.shape
    #     enc_ftrs = torchvision.transforms.CenterCrop([H, W])(enc_ftrs)
    #     return enc_ftrs


class UNet3D(nn.Module):
    def __init__(self, in_ch=1, enc_chs=(32, 64, 128, 256, 512), dec_chs=(512, 256, 128, 64, 32), out_ch=1):
        super().__init__()
        self.encoder = Encoder(in_ch, enc_chs)
        self.decoder = Decoder(dec_chs)
        self.head = nn.Conv3d(dec_chs[-1], out_ch, 1)

    def forward(self, x):
        enc_ftrs = self.encoder(x)
        out = self.decoder(enc_ftrs[::-1][0], enc_ftrs[::-1][1:])
        out = self.head(out)
        # activation = nn.Sigmoid()
        activation = nn.ReLU()
        out = activation(out)
        return out


# Defines DVH loss class
class DVHLoss(nn.Module):
    def __init__(self):
        super(DVHLoss, self).__init__()
        self.loss = torch.nn.MSELoss()

    def __call__(self, predicted_dose, target_hist, target_bins, oar):
        """
        Calculate DVH loss: averaged over all OARs. Target hist is already computed
            predicted dose (tensor) -- [N, C, D, H, W] C = 1
            target hist (tensor)    -- [N, n_bins, n_oars]
            target bins (tensor)    -- [N, n_bins]
            oar (tensor)            -- [N, C, D, H, W] C == n_oars one hot encoded OAR including PTV
        """

        # Calculate predicted hist
        vols = torch.sum(oar, axis=(2, 3, 4))
        n_bins = target_bins.shape[1]
        hist = torch.zeros_like(target_hist)
        bin_w = target_bins[0, 1] - target_bins[0, 0]

        # print(vols.shape, hist.shape)

        for i in range(n_bins):
            diff = torch.sigmoid((predicted_dose - target_bins[:, i]) / bin_w)
            # print(diff.shape)
            # diff = torch.cat(oar.shape[1] * [diff.unsqueeze(axis=1)]) * oar
            diff = diff.repeat(1, oar.shape[1], 1, 1, 1) * oar
            num = torch.sum(diff, axis=(2, 3, 4))
            # print(diff.shape, num.shape)
            hist[:, i] = (num / vols)

        # print(hist.detach().cpu().numpy())
        # print(vols.detach().cpu().numpy())

        return self.loss(hist, target_hist)


# Defines Bhattacharya loss class
class BhattLoss(nn.Module):
    def __init__(self):
        super(BhattLoss, self).__init__()
        self.loss = torch.nn.MSELoss()

    def __call__(self, predicted_dose, target_hist, target_bins, oar):
        """
        Calculate DVH loss: averaged over all OARs. Target hist is already computed
            predicted dose (tensor) -- [N, C, D, H, W] C = 1
            target hist (tensor)    -- [N, n_bins, n_oars]
            target bins (tensor)    -- [N, n_bins]
            oar (tensor)            -- [N, C, D, H, W] C == n_oars one hot encoded OAR including PTV
        """

        # Calculate predicted hist
        vols = torch.sum(oar, axis=(2, 3, 4))
        n_bins = target_bins.shape[1]
        hist = torch.zeros_like(target_hist)
        bin_w = target_bins[0, 1] - target_bins[0, 0]

        # print(vols.shape, hist.shape)

        for i in range(n_bins):
            diff = torch.sigmoid((predicted_dose - target_bins[:, i]) / bin_w)
            # print(diff.shape)
            # diff = torch.cat(oar.shape[1] * [diff.unsqueeze(axis=1)]) * oar
            diff = diff.repeat(1, oar.shape[1], 1, 1, 1) * oar
            num = torch.sum(diff, axis=(2, 3, 4))
            # print(diff.shape, num.shape)
            hist[:, i] = (num / vols)

        print(hist.detach().cpu().numpy())
        print(vols.detach().cpu().numpy())
        histprod = torch.sqrt(hist * target_hist)
        # histprod = torch.clamp(histprod, 1e-8)
        print(histprod.detach().cpu().numpy())
        bhattdist = torch.sum(histprod, axis=(1, 2))  # Sum of bhattacharya distances for each OAR
        bhattdist = torch.clamp(bhattdist, 1e-3)
        print(bhattdist.detach().cpu().numpy())
        bhattloss = torch.mean(-torch.log(bhattdist))
        return bhattloss


##Moment loss
class MomentLoss(nn.Module):
    def __init__(self):
        super(MomentLoss, self).__init__()
        self.loss = torch.nn.MSELoss()

    def __call__(self, predicted_dose, oar, dose):
        """
        Calculate DVH loss: averaged over all OARs. Target hist is already computed
            predicted dose (tensor) -- [N, C, D, H, W] C = 1
            target hist (tensor)    -- [N, n_bins, n_oars]
            target bins (tensor)    -- [N, n_bins]
            oar (tensor)            -- [N, C, D, H, W] C == n_oars one hot encoded OAR including PTV
        """

        # Calculate predicted hist
        vols = torch.sum(oar, axis=(2, 3, 4))
        # n_bins = target_bins.shape[1]
        # hist = torch.zeros_like(target_hist)
        # bin_w = target_bins[0,1] - target_bins[0,0]
        keys = ['Eso', 'Cord', 'Heart', 'Lung_L', 'Lung_R', 'PTV']
        # moment = [[1, 2, 10], [1, 2, 10], [1, 2, 10], [1, 2, 10], [1, 2, 10], [2, 4, 6]]
        # momentOfStructure = dict(zip(keys, moment))
        momentOfStructure = {'Eso': {'moments': [1, 2, 10], 'weights': [1, 1, 1]},
                             'Cord': {'moments': [1, 2, 10], 'weights': [1, 1, 1]},
                             'Heart': {'moments': [1, 2, 10], 'weights': [1, 1, 1]},
                             'Lung_L': {'moments': [1, 2, 10], 'weights': [1, 1, 1]},
                             'Lung_R': {'moments': [1, 2, 10], 'weights': [1, 1, 1]},
                             'PTV': {'moments': [2, 4, 6], 'weights': [1, 1, 1]}}
        # momentOfStructure = dict([(k, v) for k in keys for v in values])
        oarPredMoment = []
        oarRealMoment = []
        pres = 60
        epsilon = 0.00001  # Added epsilon as the loss function can become sqrt(0)
        for i in range(oar.shape[1]):
            moments = momentOfStructure[keys[i]]['moments']
            weights = momentOfStructure[keys[i]]['weights']
            for j in range(len(moments)):
                gEUDa = moments[j]
                weight = weights[j]
                oarpreddose = predicted_dose * oar[:, i, :, :, :]
                oarRealDose = dose * oar[:, i, :, :, :]
                if i < (oar.shape[1] - 1):
                    # oarPredMomenta = torch.pow((1 / vols[0, i]) * (torch.sum(torch.pow(oarpreddose-pres, gEUDa), axis=(2, 3, 4))) + epsilon, 1 / gEUDa)
                    # oarRealMomenta = torch.pow((1 / vols[0, i]) * (torch.sum(torch.pow(oarRealDose - pres, gEUDa), axis=(2, 3, 4))) + epsilon, 1 / gEUDa)
                    # print(oarPredMomenta)
                    # Use (1 / (oar[:, i, ...] == 1).sum()) to count the number of voxels
                    # else:
                    oarPredMomenta = weight * torch.pow((1 / vols[0, i]) * (torch.sum(torch.pow(oarpreddose, gEUDa), axis=(2, 3, 4))) + epsilon, 1 / gEUDa)
                    oarRealMomenta = weight * torch.pow((1 / vols[0, i]) * (torch.sum(torch.pow(oarRealDose, gEUDa), axis=(2, 3, 4))) + epsilon, 1 / gEUDa)
                    oarPredMoment.append(oarPredMomenta)
                    oarRealMoment.append(oarRealMomenta)
        oarPredMoment = torch.stack(oarPredMoment)
        oarRealMoment = torch.stack(oarRealMoment)
        # print(oarPredMoment)
        # oarPredMoment = oarPredMoment + oarPredMomenta

        return self.loss(oarPredMoment, oarRealMoment)  # self.loss(oarPredMoment, oarRealMoment)
        # #self.loss(oarPredMoment, torch.zeros_like(oarPredMoment))
        # #torch.sum(oarPredMoment)#, torch.zeros_like(oarPredMoment))


# from .load_data import load_data
# from .load_metadata import load_metadata
from .save_nrrd import save_nrrd
from .save_or_load_pickle import *
from .get_eclipse_fluence import get_eclipse_fluence
from .view_in_slicer_jupyter import view_in_slicer_jupyter
from .convert_dose_rt_dicom_to_portpy import convert_dose_rt_dicom_to_portpy



from __future__ import annotations
import os
import numpy as np
from typing import List
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from portpy.photon.plan import Plan


def get_eclipse_fluence(my_plan: Plan, sol: dict, path: str = None, beam_ids: List[str] = None) -> None:
    """
    save eclipse fluence in the path directory

    :param my_plan: object of class Plan
    :param sol: dictionary containing optimal intensity
    :param path: directory for saving the optimal fluence
    :param beam_ids: list of string containing beam ids

    """
    if path is None:
        path = os.getcwd()
    if not os.path.exists(path):
        os.makedirs(path)
    tol = 1e-06
    inf_matrix = sol['inf_matrix']
    optimal_fluence_2d = inf_matrix.fluence_1d_to_2d(sol=sol)
    for i in range(len(optimal_fluence_2d)):
        if beam_ids is not None:
            beam_id = beam_ids[i]
        else:
            beam_id = str(inf_matrix.beamlets_dict[i]['beam_id'])
        file_name = 'ID' + beam_id + '.optimal_fluence'
        filepath = os.path.join(path, file_name)
        f = open(filepath, 'w')
        fluence_2d = optimal_fluence_2d[i]/(my_plan.get_prescription()/my_plan.get_num_of_fractions())
        f.write('optimalfluence\n')
        f.write('SizeX      {}\n'.format(fluence_2d.shape[1]))
        f.write('SizeY      {}\n'.format(fluence_2d.shape[0]))
        f.write('SpacingX   {}\n'.format(2.5))
        f.write('SpacingY   {}\n'.format(2.5))
        beamlets = inf_matrix._beams.beams_dict['beamlets'][i]
        x_positions = beamlets['position_x_mm'][0] - beamlets['width_mm'][
            0] / 2  # x position is center of beamlet. Get left corner
        y_positions = beamlets['position_y_mm'][0] + beamlets['height_mm'][
            0] / 2  # y position is center of beamlet. Get top corner
        f.write('OriginX    {}\n'.format(np.min(x_positions) + 1.25))  # originX should be the center of the first beamlet, but beamlet.X is the left
        f.write('OriginY    {}\n'.format(np.max(y_positions) - 1.25))  # originY should be the center of the first beamlet, but beamlet.Y is the top
        f.write('Values\n')
        f.close()
        fluence_2d[fluence_2d < tol] = 0
        mat = np.matrix(fluence_2d)
        with open(filepath, 'a') as f:
            for line in mat:
                np.savetxt(f, line, fmt='%.6f', delimiter='\t')

        # np.savetxt(fileName, optimal_fluence_2d[i], '-append', 'delimiter'='\t', 'precision', '%.10G')




from .data_explorer import DataExplorer
from typing import List


class CT:
    """
    Class representing CT for the patient

    - **Attributes** ::

        :param ct_dict: dictionary containing ct metadata and data
        :type patient_id: patient id of the patient


    """

    def __init__(self, data: DataExplorer):
        metadata = data.load_metadata()
        ct_dict = data.load_data(meta_data=metadata['ct'])
        self.ct_dict = ct_dict
        self.patient_id = data.patient_id

    def get_ct_res_xyz_mm(self) -> List[float]:
        """

        :return: number of fractions to be delivered
        """
        return self.ct_dict['resolution_xyz_mm']



from setuptools import setup, find_packages


def _get_portpy_photon_version():
    with open('portpy/__init__.py') as fp:
        for line in fp:
            if line.startswith('__version__'):
                g = {}
                exec(line, g)  # pylint: disable=exec-used
                return g['__version__']
        raise ValueError('`__version__` not defined in `portpy/__init__.py`')


_VERSION = _get_portpy_photon_version()
STATUSES = [
    "1 - Planning",
    "2 - Pre-Alpha",
    "3 - Alpha",
    "4 - Beta",
    "5 - Production/Stable",
    "6 - Mature",
    "7 - Inactive"
]

setup(
    name='portpy',
    version=_VERSION,
    url='https://github.com/PortPy-Project/PortPy',
    license='Apache License, Version 2.0',
    packages=find_packages(exclude=["examples*"]),
    author='Gourav Jhanwar, Mojtaba Tefagh, Vicki Taasti, Seppo Tuomaala, Saad Nadeem, Masoud Zarepisheh',
    author_email="jhanwarg@mskcc.struct, mtefagh@acm.struct, vicki.taasti@maastro.nl, tuomaals@mskcc.struct, nadeems@mskcc.struct, zarepism@mskcc.struct",
    description='First open-source radiation treatment planning system in Python',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    python_requires='>=3.7',
    include_package_data=True,
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries',
    ],
    install_requires=[
        "cvxpy>=1.1.18",
        "ecos>=2.0.10",
        "h5py>=3.6.0",
        "Mosek>=9.3.14",
        "natsort>=8.1.0",
        "numpy>=1.15",
        "osqp>=0.4.1",
        "pandas>=1.1.5",
        "python-dateutil>=2.8.2",
        "pytz>=2022.1",
        "qdldl>=0.1.5",
        "scipy>=1.5.4",
        "scs>=3.2.0",
        "six>=1.16.0",
        "matplotlib>=3.5.3",
        "Shapely>=1.8.4",
        "SimpleITK>=2.0.2",
        "tabulate>=0.9.0",
        "typing>=3.7.4.3",
        "jinja2>=3.0.1",
        "typing-extensions>=3.10.0.0",
        "scikit-image>=0.17.2",
        "patchify>=0.2.3",
        "pydicom>=2.2.0",
    ],

)



from __future__ import annotations
import os
import pickle
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from portpy.photon.plan import Plan


def save_plan(my_plan: Plan, plan_name: str = None, path: str = None) -> None:
    """

    Save pickled file for plan object

    :param my_plan: object fo class Plan
    :param plan_name: create the name of the pickled file of plan object. If none, it will save with the name as 'my_plan'
    :param path: if path is set, plan object will be pickled and saved in path directory else it will save in current project directory
    :return: save pickled object of class Plan

    :Example:
    >>> my_plan.save_plan(plan_name='my_plan', path=r"path/to/save_plan")
    """
    if path is None:
        path = os.getcwd()
    elif not os.path.exists(path):
        os.makedirs(path)

    if plan_name is None:
        plan_name = 'my_plan'
    with open(os.path.join(path, plan_name), 'wb') as pickle_file:
        # pickle the dictionary and write it to file
        pickle.dump(my_plan, pickle_file)


def load_plan(plan_name: str = None, path: str = None):
    """
    Load pickle file of the plan object.

    :param plan_name: plan_name of the object of class Plan. It None, it will try to look for plan name called 'my_plan'
    :param path: if path is set, plan object will be load from path directory else current project directory
    :return: load pickled object of class Plan

    :Example:
    >>> load_plan(plan_name='my_plan', path=r"path/for/loading_plan")
    """
    if path is None:
        path = os.getcwd()

    if plan_name is None:
        plan_name = 'my_plan'
    with open(os.path.join(path, plan_name), 'rb') as pickle_file:
        my_plan = pickle.load(pickle_file)
    return my_plan


def load_optimal_sol(sol_name: str, path: str = None) -> dict:
    """

    Load optimal solution dictionary got from optimization

    :param sol_name: name of the optimal solution to be loaded.
    :param path: if path is set, plan object will be load from path directory else current directory
    :return: load solution

    :Example:
    >>> sol = load_optimal_sol(sol_name='sol', path=r'path/for/loading_sol')
    """
    if path is None:
        path = os.getcwd()

    with open(os.path.join(path, sol_name), 'rb') as pickle_file:
        sol = pickle.load(pickle_file)
    return sol


def save_optimal_sol(sol: dict, sol_name: str, path: str = None) -> None:
    """
    Save the optimal solution dictionary from optimization

    :param sol: optimal solution dictionary
    :param sol_name: name of the optimal solution saved
    :param path: if path is set, plan object will be load from path directory else current directory
    :return: save pickled file of optimal solution dictionary

    :Example:
    >>> save_optimal_sol(sol=sol, sol_name='sol', path=r'path/to/save_solution')
    """
    if path is None:
        path = os.getcwd()
    elif not os.path.exists(path):
        os.makedirs(path)
    with open(os.path.join(path, sol_name), 'wb') as pickle_file:
        pickle.dump(sol, pickle_file, protocol=4)


def save_inf_matrix(inf_matrix, inf_name: str = None, path: str = None) -> None:
    """

    Save pickled file for plan object

    :param inf_matrix: object fo class Infuence matrix
    :param inf_name: create the name of the pickled file of InfluenceMatrix object. If none, it will save with the name as 'inf_matrix'
    :param path: if path is set, plan object will be pickled and saved in path directory else it will save in current project directory
    :return: save pickled object of class Plan

    :Example:
    >>> save_inf_matrix(inf_matrix=inf_matrix, inf_name='inf_matrix', path=r"path/to/save_inf_matrix")
    """
    if path is None:
        path = os.getcwd()
    elif not os.path.exists(path):
        os.makedirs(path)

    if inf_name is None:
        inf_matrix = 'inf_matrix'
    with open(os.path.join(path, inf_name), 'wb') as pickle_file:
        # pickle the dictionary and write it to file
        pickle.dump(inf_matrix, pickle_file)


def load_inf_matrix(inf_name: str = None, path: str = None):
    """
    Load pickle file of the plan object.

    :param inf_name: influence matrix name of the object of class InfleunceMatrix.
    :param path: if path is set, plan object will be load from path directory else current project directory
    :return: load pickled object of class Plan

    :Example:
    >>> load_plan(plan_name='my_plan', path=r"path/for/loading_inf_matrix")
    """
    if path is None:
        path = os.getcwd()

    if inf_name is None:
        inf_name = 'inf_matrix'
    with open(os.path.join(path, inf_name), 'rb') as pickle_file:
        inf_matrix = pickle.load(pickle_file)
    return inf_matrix
